///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: _internals.h,v 1.19 2001/11/10 12:52:58 epocman Exp $
//	Subject : IBPP internal declarations
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* 'Internal declarations' means everything used to implement ibpp. This
//	  file and its contents is NOT needed by users of the library. All those
//	  declarations are wrapped in a namespace : 'ibpp_internals'.
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

/*
#ifndef __INTERNALS_H__
#define __INTERNALS_H__
*/

#include "ibase.h"		// From InterBase 6.0 SDK
#include "time.h"		// From C standard library

#ifdef IBPP_LINUX

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

typedef long long __int64;
typedef unsigned char BYTE;
typedef unsigned short WORD;

typedef struct _SYSTEMTIME {
	WORD wYear;
	WORD wMonth;
	WORD wDayOfWeek;
	WORD wDay;
	WORD wHour;
	WORD wMinute;
	WORD wSecond;
	WORD wMilliseconds;
} SYSTEMTIME;

void GetLocalTime (SYSTEMTIME* st);

#endif

#ifdef _DEBUG
#define ASSERTION(x)	{if (!(x)) {throw ExceptionImpl("ASSERTION", \
							"'"#x"' is not verified at %s, line %d", \
								__FILE__, __LINE__);}}
#else
#define ASSERTION(x)	/* x */
#endif

namespace ibpp_internals
{

class DatabaseImpl;
class TransactionImpl;
class StatementImpl;
class BlobImpl;
class ArrayImpl;

//	Native data types
typedef enum {ivArray, ivBlob, ivDate, ivTime, ivTimestamp, ivString,
			ivShort, ivLong, ivLarge, ivFloat, ivDouble,
			ivBool, ivDBKey} IITYPE;

//
//	Those are the Interbase C API prototypes that we use
//	Taken 'asis' from IBASE.H, prefix 'isc_' replaced with 'proto_',
//	and 'typedef' preprended...
//

typedef ISC_STATUS ISC_EXPORT proto_attach_database (ISC_STATUS ISC_FAR *, 
					    short, 
					    char ISC_FAR *, 
					    isc_db_handle ISC_FAR *, 
					    short, 
					    char ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_detach_database (ISC_STATUS ISC_FAR *,  
					    isc_db_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_drop_database (ISC_STATUS ISC_FAR *,  
					  isc_db_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_database_info (ISC_STATUS ISC_FAR *, 
					  isc_db_handle ISC_FAR *, 
					  short, 
					  char ISC_FAR *, 
					  short, 
					  char ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_execute_immediate (ISC_STATUS ISC_FAR *, 
						   isc_db_handle ISC_FAR *, 
						   isc_tr_handle ISC_FAR *, 
						   unsigned short, 
						   char ISC_FAR *, 
						   unsigned short, 
						   XSQLDA ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_open_blob2 (ISC_STATUS ISC_FAR *, 
				       isc_db_handle ISC_FAR *, 
				       isc_tr_handle ISC_FAR *,
				       isc_blob_handle ISC_FAR *, 
				       ISC_QUAD ISC_FAR *, 
				       short,  
				       char ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_create_blob (ISC_STATUS ISC_FAR *, 
					isc_db_handle ISC_FAR *, 
					isc_tr_handle ISC_FAR *, 
					isc_blob_handle ISC_FAR *, 
					ISC_QUAD ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_close_blob (ISC_STATUS ISC_FAR *, 
				       isc_blob_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_cancel_blob (ISC_STATUS ISC_FAR *, 
				        isc_blob_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_get_segment (ISC_STATUS ISC_FAR *, 
				        isc_blob_handle ISC_FAR *, 
				        unsigned short ISC_FAR *, 
				        unsigned short, 
				        char ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_put_segment (ISC_STATUS ISC_FAR *, 
					isc_blob_handle ISC_FAR *, 
					unsigned short, 
					char ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_blob_info (ISC_STATUS ISC_FAR *, 
				      isc_blob_handle ISC_FAR *, 
				      short,
 				      char ISC_FAR *, 
				      short, 
				      char ISC_FAR *);

typedef ISC_LONG    ISC_EXPORT proto_vax_integer (char ISC_FAR *, 
					short);

typedef ISC_LONG    ISC_EXPORT proto_sqlcode (ISC_STATUS ISC_FAR *);

typedef void        ISC_EXPORT proto_sql_interprete (short, 
					   char ISC_FAR *, 
					   short);

typedef ISC_STATUS  ISC_EXPORT proto_interprete (char ISC_FAR *, 
				       ISC_STATUS ISC_FAR * ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_que_events (ISC_STATUS ISC_FAR *, 
				       isc_db_handle ISC_FAR *, 
				       ISC_LONG ISC_FAR *, 
				       short, 
				       char ISC_FAR *, 
				       isc_callback, 
				       void ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_cancel_events (ISC_STATUS ISC_FAR *, 
					  isc_db_handle ISC_FAR *, 
					  ISC_LONG ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_start_multiple (ISC_STATUS ISC_FAR *, 
					   isc_tr_handle ISC_FAR *, 
					   short, 
					   void ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_commit_transaction (ISC_STATUS ISC_FAR *, 
					       isc_tr_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_commit_retaining (ISC_STATUS ISC_FAR *, 
					     isc_tr_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_rollback_transaction (ISC_STATUS ISC_FAR *, 
						 isc_tr_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_rollback_retaining (ISC_STATUS ISC_FAR *, 
						 isc_tr_handle ISC_FAR *);

///////////
typedef ISC_STATUS  ISC_EXPORT proto_dsql_allocate_statement (ISC_STATUS ISC_FAR *, 
						    isc_db_handle ISC_FAR *, 
						    isc_stmt_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_describe (ISC_STATUS ISC_FAR *, 
					  isc_stmt_handle ISC_FAR *, 
					  unsigned short, 
					  XSQLDA ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_describe_bind (ISC_STATUS ISC_FAR *, 
					       isc_stmt_handle ISC_FAR *, 
					       unsigned short, 
					       XSQLDA ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_execute (ISC_STATUS ISC_FAR *, 
					 isc_tr_handle ISC_FAR *,
					 isc_stmt_handle ISC_FAR *, 
					 unsigned short, 
					 XSQLDA ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_execute2 (ISC_STATUS ISC_FAR *, 
					  isc_tr_handle ISC_FAR *,
					  isc_stmt_handle ISC_FAR *, 
					  unsigned short, 
					  XSQLDA ISC_FAR *,
					  XSQLDA ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_fetch (ISC_STATUS ISC_FAR *, 
				       isc_stmt_handle ISC_FAR *, 
				       unsigned short, 
				       XSQLDA ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_free_statement (ISC_STATUS ISC_FAR *, 
						isc_stmt_handle ISC_FAR *, 
						unsigned short);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_prepare (ISC_STATUS ISC_FAR *, 
					 isc_tr_handle ISC_FAR *, 
					 isc_stmt_handle ISC_FAR *, 
					 unsigned short, 
					 char ISC_FAR *, 
					 unsigned short, 
				 	 XSQLDA ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_set_cursor_name (ISC_STATUS ISC_FAR *, 
						 isc_stmt_handle ISC_FAR *, 
						 char ISC_FAR *, 
						 unsigned short);

typedef ISC_STATUS  ISC_EXPORT proto_dsql_sql_info (ISC_STATUS ISC_FAR *, 
					  isc_stmt_handle ISC_FAR *, 
					  short, 
					  char ISC_FAR *, 
					  short, 
					  char ISC_FAR *);

typedef void        ISC_EXPORT proto_decode_date (ISC_QUAD ISC_FAR *, 
					void ISC_FAR *);

typedef void        ISC_EXPORT proto_encode_date (void ISC_FAR *, 
					ISC_QUAD ISC_FAR *);

typedef int			ISC_EXPORT proto_add_user (ISC_STATUS ISC_FAR *, USER_SEC_DATA *);
typedef int			ISC_EXPORT proto_delete_user (ISC_STATUS ISC_FAR *, USER_SEC_DATA *);
typedef int			ISC_EXPORT proto_modify_user (ISC_STATUS ISC_FAR *, USER_SEC_DATA *);

//
//	Those API are only available in versions 6.x of the GDS32.DLL
//

typedef ISC_STATUS  ISC_EXPORT proto_service_attach (ISC_STATUS ISC_FAR *, 
					   unsigned short, 
					   char ISC_FAR *,
					   isc_svc_handle ISC_FAR *, 
					   unsigned short, 
					   char ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_service_detach (ISC_STATUS ISC_FAR *, 
					   isc_svc_handle ISC_FAR *);

typedef ISC_STATUS  ISC_EXPORT proto_service_query (ISC_STATUS ISC_FAR *, 
					  isc_svc_handle ISC_FAR *,
                      		          isc_resv_handle ISC_FAR *,
					  unsigned short, 
					  char ISC_FAR *, 
					  unsigned short, 
					  char ISC_FAR *, 
					  unsigned short, 
					  char ISC_FAR *);

typedef ISC_STATUS ISC_EXPORT proto_service_start (ISC_STATUS ISC_FAR *,
    					 isc_svc_handle ISC_FAR *,
                         		 isc_resv_handle ISC_FAR *,
    					 unsigned short,
    					 char ISC_FAR*);

typedef void        ISC_EXPORT proto_decode_sql_date (ISC_DATE ISC_FAR *, 
					void ISC_FAR *);

typedef void        ISC_EXPORT proto_decode_sql_time (ISC_TIME ISC_FAR *, 
					void ISC_FAR *);

typedef void        ISC_EXPORT proto_decode_timestamp (ISC_TIMESTAMP ISC_FAR *, 
					void ISC_FAR *);

typedef void        ISC_EXPORT proto_encode_sql_date (void ISC_FAR *, 
					ISC_DATE ISC_FAR *);

typedef void        ISC_EXPORT proto_encode_sql_time (void ISC_FAR *, 
					ISC_TIME ISC_FAR *);

typedef void        ISC_EXPORT proto_encode_timestamp (void ISC_FAR *, 
					ISC_TIMESTAMP ISC_FAR *);

//
//	Internal binding structure to the GDS32 DLL
//

struct GDS
{
	static GDS* Instance;
	static GDS* gds(void);
	static void Release(void);

	// Attributes
#ifdef IBPP_WINDOWS
	HMODULE _Handle;		// The GDS32.DLL HMODULE, if it could be loaded
#endif	
	int _GDSVersion; 		// Version of the GDS32.DLL (50 for 5.0, 60 for 6.0)

	// GDS32 Entry Points
	proto_attach_database*			_attach_database;
	proto_detach_database*			_detach_database;
	proto_drop_database*			_drop_database;
	proto_database_info*			_database_info;
	proto_dsql_execute_immediate*	_dsql_execute_immediate;
	proto_open_blob2*				_open_blob2;
	proto_create_blob*				_create_blob;
	proto_close_blob*				_close_blob;
	proto_cancel_blob*				_cancel_blob;
	proto_get_segment*				_get_segment;
	proto_put_segment*				_put_segment;
	proto_blob_info*				_blob_info;

	proto_vax_integer*				_vax_integer;
	proto_sqlcode*					_sqlcode;
	proto_sql_interprete*			_sql_interprete;
	proto_interprete*				_interprete;
	proto_que_events*				_que_events;
	proto_cancel_events* 			_cancel_events;
	proto_start_multiple*			_start_multiple;
	proto_commit_transaction*		_commit_transaction;
	proto_commit_retaining*			_commit_retaining;
	proto_rollback_transaction*		_rollback_transaction;
	proto_rollback_retaining*		_rollback_retaining;
	proto_dsql_allocate_statement*	_dsql_allocate_statement;
	proto_dsql_describe*			_dsql_describe;
	proto_dsql_describe_bind*		_dsql_describe_bind;
	proto_dsql_prepare*				_dsql_prepare;
	proto_dsql_execute*				_dsql_execute;
	proto_dsql_execute2*			_dsql_execute2;
	proto_dsql_fetch*				_dsql_fetch;
	proto_dsql_free_statement*		_dsql_free_statement;
	proto_dsql_set_cursor_name*		_dsql_set_cursor_name;
	proto_dsql_sql_info* 			_dsql_sql_info;
	proto_decode_date*				_decode_date;
	proto_encode_date*				_encode_date;
	proto_add_user*					_add_user;
	proto_delete_user*				_delete_user;
	proto_modify_user*				_modify_user;


	proto_service_attach*			_service_attach;
	proto_service_detach*			_service_detach;
	proto_service_start*			_service_start;
	proto_service_query*			_service_query;
	proto_decode_sql_date*			_decode_sql_date;
	proto_decode_sql_time*			_decode_sql_time;
	proto_decode_timestamp*			_decode_timestamp;
	proto_encode_sql_date*			_encode_sql_date;
	proto_encode_sql_time*			_encode_sql_time;
	proto_encode_timestamp*			_encode_timestamp;

protected:
	// Constructor & Destructor
	GDS() { };
	~GDS() { };
};

//
//	Service Parameter Block (used to define a service)
//

class _SPB
{
    char* _Buffer;				// Dynamically allocated SPB structure
    short _Size;  				// Its size in bytes

	void Alloc(void);

public:
    void Insert(char);			// Insert a single byte code
    void Insert(char, int, const char*);	// Insert a string, len can be defined as 1 or 2 bytes
	void Insert(char type, int data);
    void Reset(void);			// Clears the SPB
	char* Self(void) { return _Buffer; }
    short Size(void) { return _Size; }

	_SPB() { _Buffer = NULL; _Size = 0; }
    ~_SPB() { Reset(); }
};

//
//	Database Parameter Block (used to define a database)
//

class _DPB
{
    char* _Buffer;				// Dynamically allocated DPB structure
    short _Size;  				// Its size in bytes

public:
    void Insert(char, const char*);	// Insert a new char* 'cluster'
    void Insert(char, short);	// Insert a new short 'cluster'
    void Insert(char, bool);   	// Insert a new bool 'cluster'
    void Insert(char, char);   	// Insert a new byte 'cluster'
    void Reset(void);			// Clears the DPB
	char* Self(void) { return _Buffer; }
    short Size(void) { return _Size; }

	_DPB() { _Buffer = NULL; _Size = 0; }
    ~_DPB() { Reset(); }
};

//
//	Transaction Parameter Block (used to define a transaction)
//

class _TPB
{
    char* _Buffer;					// Dynamically allocated TPB structure
    short _Size;	   				// Its size in bytes

public:
    void Insert(char);				// Insert a flag item
    void Insert(char, const char*);		// Insert a flag+string item
    void Reset(void);				// Clears the TPB
	char* Self(void) { return _Buffer; }
    short Size(void) { return _Size; }

	_TPB() { _Buffer = NULL; _Size = 0; }
    ~_TPB() { Reset(); }
};

//
//	Events Handling. EPB stands for Events Paramater Block, but it really
//	encapsulate all the management of a list of events to wait for.
//

class _EPB
{
    char* _EvnBuffer;				// Dynamically allocated buffer
    char* _ResBuffer;				// Dynamically allocated buffer
    short _BufSize;	   				// Buffers used size
	short _BufAlloc;				// Buffers allocated size

	IBPP::EventInterface** _ObjRefs;// Array of object references
	short _EvnCount;				// Count of events defined
	short _EvnAlloc;				// Count of events allocated in arrays

public:
    void Define(const char*, IBPP::EventInterface*);// Define (add) an event to the list
    void Drop(const char*);					// Drop an event from the list
	void FireActions(IBPP::Database*);
	char* EventsBuffer(void) { return _EvnBuffer; }
	char* ResultsBuffer(void) { return _ResBuffer; }
    short Size(void) { return _BufSize; }

	_EPB();
    ~_EPB();
};

//
//	Used to receive (and process) a results buffer in various API calls
//

class _RB
{
	char* _Buffer;
	short _Size;

	char* FindToken(char token);
	char* FindToken(char token, char subtoken);

public:
	void Reset(void);
	int GetValue(char token);
	int GetValue(char token, char subtoken);
	bool GetBool(char token);
	int GetString(char token, char* data);

	char* Self(void) { return _Buffer; }
	short Size(void) { return _Size; }

	_RB();
	_RB(int Size);
	~_RB();
};

class _XSQLDA
{
	XSQLDA* _DescrArea;				// XSQLDA descriptor itself
	double* _Numerics;				// Temporary storage for Numerics
	float* _Floats;					// Temporary storage for Floats
	__int64* _Larges;				// Temporary storage for Larges
	long* _Longs;					// Temporary storage for Longs
	short* _Shorts;					// Temporary storage for Shorts
	bool* _Bools;					// Temporary storage for Bools

public:
	void Free(void);
	short AllocatedSize(void) { return _DescrArea->sqln; }
	short DescrSize(void) { return _DescrArea->sqld; }
	void Resize(short n);
	void AllocVariables(void);
	void SetNull(int);
	bool IsNull(int);
	void SetValue(int, IITYPE, void* value);
	void* GetValue(int, IITYPE, void* = NULL);
	const char* GetName(int);
	const char* GetAlias(int);
	const char* GetTable(int);
	int GetVarNum(const char*);
	int GetVarNumAlias(const char*);
	IBPP::SDT GetType(int);
	int GetSize(int);
	int GetScale(int);
	XSQLDA* Self(void) { return _DescrArea; }

	// Constructors and Destructors
	_XSQLDA();
	_XSQLDA(const _XSQLDA&);
	_XSQLDA(short n);
    ~_XSQLDA();
};

//
//	Though strictly internal, this class is not named with the "Impl" suffix.
//	This is done because there is NO _Status public class and so this one is
//	not the "hidden" implementation of the public one.
//

class _IBS
{
	ISC_STATUS _Vector[20];
	char* _Message;

public:
	ISC_STATUS* Self(void) { return _Vector; }
	bool Errors(void) { return (_Vector[0] == 1 && _Vector[1]) ? true : false; }
    void Printf(void);
	const char* ErrorMessage(void);
	int SqlCode(void);
	void Reset(void);

	_IBS();
	_IBS(_IBS&);	// Copy Constructor
	~_IBS();
};

///////////////////////////////////////////////////////////////////////////////
//
//	Implementation of the "hidden" classes associated with their public
//	counterparts. Their private data and methods can freely change without
//	breaking the compatibility of the DLL. If they receive new public methods,
//	and those methods are reflected in the public class, then the compatibility
//	is broken.
//
///////////////////////////////////////////////////////////////////////////////

//
//	Hidden implementation of Exception user class.
//	The implementation sources never throw this internal exception class.
//	Instead they directly throw the public version (IException) itself.
//	This eliminates the need to try - catch every public method call just to
//	transform an internal exception object in a public one.
//

class ExceptionImpl : public IBPP::Exception
{
	//	(((((((( OBJECT INTERNALS ))))))))

private:
	_IBS* _Status;
	char* _Origin;				// Exception Origin ("IDatabase::Drop")
	char* _Message;				// Message attached to the exception

	void BlindCopy(const ExceptionImpl&);
	void SetStatus(_IBS* status);
	void SetOrigin(const char* origin);
	void SetMessage(const char* message);

public:
	ExceptionImpl();
	ExceptionImpl(const ExceptionImpl&);
	ExceptionImpl& operator=(const ExceptionImpl&);
	ExceptionImpl(const char* origin, const char* message = NULL, ...);
	ExceptionImpl(_IBS* Status, const char* origin, const char* message = NULL);
	virtual ~ExceptionImpl();

	//	(((((((( OBJECT INTERFACE ))))))))

public:
    const char* GetOrigin(void);
    void GetErrorMessage(char*);
	int SqlCode(void) { return _Status != NULL ? _Status->SqlCode() : -999; }
	void PrintfErrorMessage(void);
};

class ExceptionHelper : public IBPP::Exception
{
	//	(((((((( OBJECT INTERNALS ))))))))

private:
	char* _Origin;				// Exception Origin ("IDatabase::Drop")
	char* _Message;				// Message attached to the exception

	void BlindCopy(const ExceptionHelper&);
	void SetOrigin(const char* origin);
	void SetMessage(const char* message);

public:
	ExceptionHelper();
	ExceptionHelper(const ExceptionHelper&);
	ExceptionHelper& operator=(const ExceptionHelper&);
	ExceptionHelper(const char* origin, const char* message = NULL);
	virtual ~ExceptionHelper();

	//	(((((((( OBJECT INTERFACE ))))))))

public:
	const char* GetOrigin(void);
	void GetErrorMessage(char*);
	int SqlCode(void) { return -999; }
	void PrintfErrorMessage(void);
};

class ServiceImpl : public IBPP::Service
{
	//	(((((((( OBJECT INTERNALS ))))))))

private:
    isc_svc_handle _Handle;		// InterBase API Service Handle
	char* _ServerName;			// Nom du serveur
    char* _UserName;			// Nom de l'utilisateur
    char* _UserPassword;		// Mot de passe de l'utilisateur

	isc_svc_handle* GetHandlePtr(void) { return &_Handle; }
	void SetServerName(const char*);
	void SetUserName(const char*);
	void SetUserPassword(const char*);

public:
	void* GetHandle(void) { return (void*)_Handle; }

	ServiceImpl(const char* ServerName, const char* UserName, const char* UserPassword);
    ~ServiceImpl();

	//	(((((((( OBJECT INTERFACE ))))))))

public:
    void Connect(void);
	void Disconnect(void);

	void AddUser(const char* username, const char* password,
			const char* first, const char* middle, const char* last);
	void ModifyUser(const char* username, const char* password,
			const char* first, const char* middle, const char* last);
	void RemoveUser(const char* username);

	void StartBackup(const char* dbfile, const char* bkfile, bool verbose);
	void StartRestore(const char* bkfile, const char* dbfile, int pagesize, bool verbose);

	bool Wait(char* line);
	bool Wait(void) { return Wait(NULL); }
	
	void Release(IBPP::Service*&);
};

class DatabaseImpl : public IBPP::Database
{
	//	(((((((( OBJECT INTERNALS ))))))))

    isc_db_handle _Handle;		// InterBase API Session Handle
	char* _ServerName;			// Nom du serveur
    char* _DatabaseName;		// Nom de la base de donnes (path/file)
    char* _UserName;			// Nom de l'utilisateur
    char* _UserPassword;		// Mot de passe de l'utilisateur
    char* _RoleName;			// Role pour tablir la connexion
	char* _CreateParams; 		// Autres paramtres (pour la cration)

	int _Dialect;					// 1 if IB5, 3 if IB6
	TransactionImpl** _Transactions;// Table of Transaction*
	StatementImpl** _Statements;	// Table of Statement*
	BlobImpl** _Blobs;				// Table of Blob*
	ArrayImpl** _Arrays;			// Table of Array*
	int _TrCount;					// Nombre de transactions associes
	int _StCount;					// Nombre de statements associes
	int _BbCount;					// Nombre Blobs attachs
	int _ArCount;					// Number of attached Arrays

	_EPB* _Events;					// Events list
	ISC_LONG _EventsId;				// IB internal Id of these events
	bool _EventsActive;				// Activated ?
	bool _EventsTrapped;			// Trapped Events ?

	void Init(void);				// Only for constructors
	void CancelEvents(void);
	void EventUpdateCounts(short size, const char* tmpbuffer);

	static void EventHandler(const char*, short, const char*);

public:
	isc_db_handle* GetHandlePtr(void) { return &_Handle; }
	void* GetHandle(void) { return (void*)_Handle; }

	DatabaseImpl(const char* ServerName, const char* DatabaseName = NULL,
				const char* UserName = NULL, const char* UserPassword = NULL,
				const char* RoleName = NULL);
    ~DatabaseImpl();

	//	(((((((( OBJECT INTERFACE ))))))))

public:
	void SetServerName(const char* ServerName);
	void SetDatabaseName(const char* DatabaseName);
	void SetUsername(const char* UserName);
	void SetUserPassword(const char* UserPassword);
	void SetRoleName(const char* RoleName);
	void SetCreateParams(const char* CreateParams);

	const char* GetServerName(void);
	const char* GetDatabaseName(void);
	const char* GetUsername(void);
	const char* GetUserPassword(void);
	const char* GetRoleName(void);
	const char* GetCreateParams(void);
	unsigned short GetDialect(void) { return (unsigned short) _Dialect; }

	void AttachTransaction(IBPP::Transaction*);
	void DetachTransaction(IBPP::Transaction*);
	void AttachStatement(IBPP::Statement*);
	void DetachStatement(IBPP::Statement*);
	void AttachBlob(IBPP::Blob*);
	void DetachBlob(IBPP::Blob*);
	void AttachArray(IBPP::Array*);
	void DetachArray(IBPP::Array*);

	void GetInfo(int* ODS, int* ODSMinor, int* PageSize, bool* Sync);
	void GetStat(int* Pages, int* Memory, int* Buffers,
		int* Fetches, int* marks, int* reads, int* writes);

    void Create(void);
	void Connect(void);
	void Disconnect(void);
    void Drop(void);

	void DefineEvent(const char*, IBPP::EventInterface*);
	void DropEvent(const char*);
	void ClearEvents(void);
	void DispatchEvents(void);

	void Release(IBPP::Database*&);
};

class TransactionImpl : public IBPP::Transaction
{
	//	(((((((( OBJECT INTERNALS ))))))))

private:
    isc_tr_handle _Handle;			// Transaction InterBase

	DatabaseImpl** _Databases;   	// Tableau de IDatabase*
	StatementImpl** _Statements;	// Tableau de IStatement*
	BlobImpl** _Blobs;				// Tableau de IBlob*
	ArrayImpl** _Arrays;			// Tableau de Array*
	_TPB** _TPBs;					// Tableau de _TPB*
	int _DbCount;		  			// Nombre d'lments de _Databases[] & _TPBs[]
	int _StCount;					// Nombre d'lments de _Statements[]
	int _BbCount;					// Nombre d'lments de _Blobs[]
	int _ArCount;					// Number of attached arrays
	int _RwCount;					// Number of attached rows

	void Init(void);			// A usage exclusif des constructeurs

public:
	isc_tr_handle* GetHandlePtr(void) { return &_Handle; }
	void* GetHandle(void) { return (void*)_Handle; }
	void AttachStatement(StatementImpl*);
	void DetachStatement(StatementImpl*);
	void AttachBlob(BlobImpl*);
	void DetachBlob(BlobImpl*);
	void AttachArray(ArrayImpl*);
	void DetachArray(ArrayImpl*);

	TransactionImpl(DatabaseImpl* db, IBPP::TAM am = IBPP::amWrite,
		IBPP::TIL il = IBPP::ilConcurrency, IBPP::TLR lr = IBPP::lrWait);
    ~TransactionImpl();

	//	(((((((( OBJECT INTERFACE ))))))))

public:
    void AttachDatabase(IBPP::Database* db, IBPP::TAM am = IBPP::amWrite,
			IBPP::TIL il = IBPP::ilConcurrency, IBPP::TLR lr = IBPP::lrWait);
    void DetachDatabase(IBPP::Database* db);
    void Start(void);
    void Commit(void);
    void Rollback(void);
    void CommitRetain(void);
	void RollbackRetain(void);

	void Release(IBPP::Transaction*&);
};

class StatementImpl : public IBPP::Statement
{
	//	(((((((( OBJECT INTERNALS ))))))))

private:
	friend TransactionImpl;

	isc_stmt_handle _Handle;	// Rfrence d'accs au statement InterBase API

	DatabaseImpl* _Database;		// Database attache
	TransactionImpl* _Transaction;	// Transaction attache
	_XSQLDA* _InDescr;			// XSQLDA pour les paramtres
	bool* _InMissing;			// Quels paramtres n'ont pas t spcifis
	_XSQLDA* _OutDescr;		// XSQLDA pour les rsultats
	bool _ResultSetAvailable;	// A t excut et un result set est disponible
	IBPP::STT _Type;			// Type de requte

	// Mthodes internes
	void Init(void);	// A usage exclusif des constructeurs
	void CursorFree(void);

public:
	// Mthodes d'accs aux proprits et attributs
	isc_stmt_handle GetHandle(void) { return _Handle; }

	StatementImpl(DatabaseImpl* database,
		TransactionImpl* = NULL, const char* = NULL);
    ~StatementImpl();

	//	(((((((( OBJECT INTERFACE ))))))))

public:
	void AttachDatabase(IBPP::Database*);
	void DetachDatabase(void);
	void AttachTransaction(IBPP::Transaction*);
	void DetachTransaction(void);
	void Prepare(const char* sql);
	void Execute(const char* sql = NULL);
	void ExecuteImmediate(const char*);
	void CursorExecute(const char* cursor, const char* sql = NULL);
	bool Fetch(void);
	int AffectedRows(void);
	void Close(void);	// Libre ressources, attachments maintenus
	IBPP::STT GetType(void) { return _Type; }

	void SetNull(int);
	void Set(int, bool);
	void Set(int, const char*);
	void Set(int, short value);
	void Set(int, int value);
	void Set(int, long value);
	void Set(int, __int64 value);
	void Set(int, float value);
	void Set(int, double value);
	void Set(int, IBPP::Timestamp& value);
	void Set(int, IBPP::Date& value);
	void Set(int, IBPP::Time& value);
	void Set(int, IBPP::DBKey& value);
	void Set(int, IBPP::Blob* value);
	void Set(int, IBPP::Array* value);

	bool IsNull(int);
	bool Get(int, bool*);
	bool Get(int, char*);
	bool Get(int, short*);
	bool Get(int, int*);
	bool Get(int, long*);
	bool Get(int, __int64*);
	bool Get(int, float*);
	bool Get(int, double*);
	bool Get(int, IBPP::Timestamp& value);
	bool Get(int, IBPP::Date& value);
	bool Get(int, IBPP::Time& value);
	bool Get(int, IBPP::DBKey& value);
	bool Get(int, IBPP::Blob* value);
	bool Get(int, IBPP::Array* value);

	bool IsNull(const char*);
	bool Get(const char*, bool*);
	bool Get(const char*, char*);
	bool Get(const char*, short*);
	bool Get(const char*, int*);
	bool Get(const char*, long*);
	bool Get(const char*, __int64*);
	bool Get(const char*, float*);
	bool Get(const char*, double*);
	bool Get(const char*, IBPP::Timestamp& value);
	bool Get(const char*, IBPP::Date& value);
	bool Get(const char*, IBPP::Time& value);
	bool Get(const char*, IBPP::DBKey& value);
	bool Get(const char*, IBPP::Blob* value);
	bool Get(const char*, IBPP::Array* value);

	int ColumnNum(const char*);
    int ColumnNumAlias(const char*);
	const char* ColumnName(int);
	const char* ColumnAlias(int);
	const char* ColumnTable(int);
	IBPP::SDT ColumnType(int);
	int ColumnSize(int);
	int ColumnScale(int);

	int Columns(void);

	void Release(IBPP::Statement*&);
};

class BlobImpl : public IBPP::Blob
{
	//	(((((((( OBJECT INTERNALS ))))))))

private:
	friend class _XSQLDA;

	bool				_IdAssigned;
	ISC_QUAD			_Id;
	isc_blob_handle		_Handle;
	bool				_WriteMode;
	DatabaseImpl*  		_Database;		// Database attache
	TransactionImpl*	_Transaction;	// Transaction attache

	void Init(void);
	void SetId(ISC_QUAD*);
	void GetId(ISC_QUAD*);

public:
	BlobImpl(const BlobImpl&);
	BlobImpl(DatabaseImpl*, TransactionImpl* = NULL);
	~BlobImpl();

	//	(((((((( OBJECT INTERFACE ))))))))

public:
	void AttachDatabase(IBPP::Database*);
	IBPP::Database* GetDatabase(void);
	void DetachDatabase(void);
	void AttachTransaction(IBPP::Transaction*);
	IBPP::Transaction* GetTransaction(void);
	void DetachTransaction(void);
	void Create(void);
	void Open(void);
	void Close(void);
	void Cancel(void);
	int Read(void*, int size);
	void Write(void*, int size);
	void GetInfo(int* Size, int* Largest, int* Segments);

	void Release(IBPP::Blob*&);
};

class ArrayImpl : public IBPP::Array
{
	//	(((((((( OBJECT INTERNALS ))))))))

private:
	friend class _XSQLDA;

	bool				_IdAssigned;
	ISC_QUAD			_Id;
	/*
	isc_blob_handle		_Handle;
	bool				_WriteMode;
	*/
	DatabaseImpl*  		_Database;		// Database attache
	TransactionImpl*	_Transaction;	// Transaction attache

	void Init(void);
	void SetId(ISC_QUAD*);
	void GetId(ISC_QUAD*);

public:
	ArrayImpl(const ArrayImpl&);
	ArrayImpl(DatabaseImpl*, TransactionImpl* = NULL);
	~ArrayImpl();

	//	(((((((( OBJECT INTERFACE ))))))))

public:
	void AttachDatabase(IBPP::Database*);
	IBPP::Database* GetDatabase(void);
	void DetachDatabase(void);
	void AttachTransaction(IBPP::Transaction*);
	IBPP::Transaction* GetTransaction(void);
	void DetachTransaction(void);
	/*
	void Create(void);
	void Open(void);
	void Close(void);
	void Cancel(void);
	int Read(void*, int size);
	void Write(void*, int size);
	void GetInfo(int* Size, int* Largest, int* Segments);
	*/
	void Release(IBPP::Array*&);
};

}	// namespace ibpp_internal

/*
#endif
*/

//
//	Eof
//
