///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: _rb.cpp,v 1.2 2001/02/01 19:18:07 epocman Exp $
//	Subject : IBPP, internal RB class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* RB == Result Block/Buffer, see Interbase 6.0 C-API
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

char* _RB::FindToken(char token)
{
	char* p = _Buffer;

	while (*p != isc_info_end)
	{
		int len;

		if (*p == token) return p;
		len = (*GDS::gds()->_vax_integer)(p+1, 2);
		p = p + len + 3;
	}

	return NULL;
}

char* _RB::FindToken(char token, char subtoken)
{
	char* p = _Buffer;

	while (*p != isc_info_end)
	{
		int len;

		if (*p == token)
		{
			// Found token, now find subtoken
			int inlen = (*GDS::gds()->_vax_integer)(p+1, 2);
			p = p + 3;
			while (inlen > 0)
			{
				if (*p == subtoken) return p;
				len = (*GDS::gds()->_vax_integer)(p+1, 2);
				p = p + len + 3;
				inlen = inlen - len - 3;
			}
			return NULL;
		}
		len = (*GDS::gds()->_vax_integer)(p+1, 2);
		p = p + len + 3;
	}

	return NULL;
}

int _RB::GetValue(char token)
{
	int value;
	int len;
	char* p = FindToken(token);

	if (p == NULL)
		throw ExceptionImpl("IIRB::GetValue", "Token not found.");

	len = (*GDS::gds()->_vax_integer)(p+1, 2);
	value = (*GDS::gds()->_vax_integer)(p+3, (short)len);

	return value;
}

int _RB::GetValue(char token, char subtoken)
{
	int value;
	int len;
	char* p = FindToken(token, subtoken);

	if (p == NULL)
		throw ExceptionImpl("IIRB::GetValue", "Token/Subtoken not found.");

	len = (*GDS::gds()->_vax_integer)(p+1, 2);
	value = (*GDS::gds()->_vax_integer)(p+3, (short)len);

	return value;
}

bool _RB::GetBool(char token)
{
	int value;
	char* p = FindToken(token);

	if (p == NULL)
		throw ExceptionImpl("IIRB::GetBool", "Token not found.");

	value = (*GDS::gds()->_vax_integer)(p+1, 4);

	return value == 0 ? false : true;
}

int _RB::GetString(char token, char* data)
{
	int len;
	char* p = FindToken(token);

	if (data == NULL)
		throw ExceptionImpl("IIRB::GetString", "NULL data pointer.");
	if (p == NULL)
		throw ExceptionImpl("IIRB::GetString", "Token not found.");

	len = (*GDS::gds()->_vax_integer)(p+1, 2);
	if (len != 0) strncpy(data, p+3, len);
	data[len] = '\0';
	return len;
}

void _RB::Reset(void)
{
	delete [] _Buffer;
	_Buffer = new char [_Size];
}

_RB::_RB()
{
	_Size = 1024;
	_Buffer = new char [1024];
}

_RB::_RB(int Size)
{
	_Size = (short)Size;
	_Buffer = new char [Size];
}

_RB::~_RB()
{
	delete [] _Buffer;
}

//
//	EOF
//
