///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: _spb.cpp,v 1.4 2001/05/22 22:26:08 epocman Exp $
//	Subject : IBPP, internal SPB class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* SPB == Service Parameter Block/Buffer, see Interbase 6.0 C-API
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>
#include <string.h>

using namespace ibpp_internals;

void _SPB::Alloc(void)
{
	if (_Buffer == NULL)
	{
		_Buffer = new char[512];
	    _Size = 0;
	}
}

void _SPB::Insert(char opcode)
{
    Alloc();
    _Buffer[_Size++] = opcode;
}

void _SPB::Insert(char type, int lenwidth, const char* data)
{
	short len = (short)strlen(data);

    Alloc();
    _Buffer[_Size++] = type;
	switch (lenwidth)
	{
		case 1 :	_Buffer[_Size] = char(len); _Size++; break;
		case 2 :	*(short*)&_Buffer[_Size] = short(len); _Size += (short)2; break;
		default :	throw ExceptionImpl("IISPB::IISPB", "Invalid length parameter");
	}
    strncpy(&_Buffer[_Size], data, len);
    _Size += len;
}

void _SPB::Insert(char type, int data)
{
    Alloc();
    _Buffer[_Size++] = type;
	*(int*)&_Buffer[_Size] = data;
	_Size += (short)4;
}

void _SPB::Reset()
{
	if (_Buffer != NULL)
    {
    	delete [] _Buffer;
        _Buffer = NULL;
        _Size = 0;
    }
}

/*
void _SPB::Insert(char type, short data)
{
	Alloc();
    _Buffer[_Size++] = type;
	_Buffer[_Size++] = char(2);
    *(short*)&_Buffer[_Size] = data;
    _Size += short(2);
}

void _SPB::Insert(char type, bool data)
{
    Alloc();
    _Buffer[_Size++] = type;
	_Buffer[_Size++] = char(1);
    _Buffer[_Size++] = char(data ? 1 : 0);
}

void _SPB::Insert(char type, char data)
{
    Alloc();
    _Buffer[_Size++] = type;
	_Buffer[_Size++] = char(1);
    _Buffer[_Size++] = data;
}
*/

//
//	EOF
//
