///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: _xsqlda.cpp,v 1.14 2001/11/03 09:36:26 epocman Exp $
//	Subject : IBPP, _XSQLDA class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* XSQLDA == See Interbase 6.0 C-API
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <math.h>
#include <time.h>
#include <stdio.h>

using namespace ibpp_internals;

#ifdef IBPP_LINUX
/* strupr is not defined in libc */
char* strupr (char* buf) 
{
	int i = 0;
	while (buf[i])
		buf[i] = toupper(buf[i]);
	return buf;
}
#endif

const double dscales[19] =
{
	1, 1E1, 1E2, 1E3, 1E4,	1E5, 1E6, 1E7, 1E8,
	1E9, 1E10, 1E11, 1E12, 1E13, 1E14, 1E15,
	1E16, 1E17, 1E18
};

const short minshort = -32768;
const short maxshort = 32767;
const long minlong = 0xFFFFFFFF;
const long maxlong = 0xEFFFFFFF;

void _XSQLDA::Free(void)
{
	if (_DescrArea != NULL)
	{
		for (int i = 0; i < _DescrArea->sqln; i++)
		{
			XSQLVAR* var = &(_DescrArea->sqlvar[i]);
			if (var->sqldata != NULL)
			{
				switch (var->sqltype & ~1)
				{
					case SQL_ARRAY :
					case SQL_BLOB :		delete (ISC_QUAD*) var->sqldata; break;
					case SQL_TIMESTAMP :	delete (ISC_TIMESTAMP*) var->sqldata; break;
					case SQL_TYPE_TIME :	delete (ISC_TIME*) var->sqldata; break;
					case SQL_TYPE_DATE :	delete (ISC_DATE*) var->sqldata; break;
					case SQL_TEXT :
					case SQL_VARYING :	delete [] var->sqldata; break;
					case SQL_SHORT :	delete (short*) var->sqldata; break;
					case SQL_LONG :		delete (long*) var->sqldata; break;
					case SQL_INT64 :	delete (__int64*) var->sqldata; break;
					case SQL_FLOAT : 	delete (float*) var->sqldata; break;
					case SQL_DOUBLE :	delete (double*) var->sqldata; break;
					default : throw ExceptionImpl("_XSQLDA::Free",
								"Found an unknown sqltype !");
				}
			}
			if (var->sqlind != NULL) delete var->sqlind;
		}
		delete [] (char*)_DescrArea;
		_DescrArea = NULL;
	}
	if (_Numerics != NULL)
	{
		delete [] _Numerics;
		_Numerics = NULL;
	}
	if (_Floats != NULL)
	{
		delete [] _Floats;
		_Floats = NULL;
	}
	if (_Larges != NULL)
	{
		delete [] _Larges;
		_Larges = NULL;
	}
	if (_Longs != NULL)
	{
		delete [] _Longs;
		_Longs = NULL;
	}
	if (_Shorts != NULL)
	{
		delete [] _Shorts;
		_Shorts = NULL;
	}
	if (_Bools != NULL)
	{
		delete [] _Bools;
		_Bools = NULL;
	}
}

void _XSQLDA::Resize(short n)
{
	int size = XSQLDA_LENGTH(n);

	Free();
    _DescrArea = (XSQLDA*) new char[size];

	memset(_DescrArea, 0, size);
	_Numerics = new double[n];
	_Floats = new float[n];
	_Larges = new __int64[n];
	_Longs = new long[n];
	_Shorts = new short[n];
	_Bools = new bool[n];
	for (int i = 0; i < n; i++)
	{
		_Numerics[i] = 0.0;
		_Floats[i] = 0.0;
		_Larges[i] = 0;
		_Longs[i] = 0;
		_Shorts[i] = 0;
		_Bools[i] = false;
	}

	_DescrArea->version = SQLDA_VERSION1;
	_DescrArea->sqln = n;
}

void _XSQLDA::AllocVariables(void)
{
	int i;
	for (i = 0; i < _DescrArea->sqld; i++)
	{
		XSQLVAR* var = &(_DescrArea->sqlvar[i]);
		switch (var->sqltype & ~1)
		{
			case SQL_ARRAY :
			case SQL_BLOB :		var->sqldata = (char*) new ISC_QUAD;
								memset(var->sqldata, 0, sizeof(ISC_QUAD));
								break;
			case SQL_TIMESTAMP :var->sqldata = (char*) new ISC_TIMESTAMP;
								memset(var->sqldata, 0, sizeof(ISC_TIMESTAMP));
								break;
			case SQL_TYPE_TIME :var->sqldata = (char*) new ISC_TIME;
								memset(var->sqldata, 0, sizeof(ISC_TIME));
								break;
			case SQL_TYPE_DATE :var->sqldata = (char*) new ISC_DATE;
								memset(var->sqldata, 0, sizeof(ISC_DATE));
								break;
			case SQL_TEXT :		var->sqldata = new char[var->sqllen+1];
								memset(var->sqldata, ' ', var->sqllen);
								var->sqldata[var->sqllen] = '\0';
								break;
			case SQL_VARYING :	var->sqldata = new char[var->sqllen+3];
								memset(var->sqldata, 0, 2);
								memset(var->sqldata+2, ' ', var->sqllen);
								var->sqldata[var->sqllen+2] = '\0';
								break;
			case SQL_SHORT :	var->sqldata = (char*) new short(0); break;
			case SQL_LONG :		var->sqldata = (char*) new long(0); break;
			case SQL_INT64 :	var->sqldata = (char*) new __int64(0); break;
			case SQL_FLOAT : 	var->sqldata = (char*) new float(0.0); break;
			case SQL_DOUBLE :	var->sqldata = (char*) new double(0.0); break;
			default : throw ExceptionImpl("_XSQLDA::AllocVariables",
						"Found an unknown sqltype !");
		}
		if (var->sqltype & 1) var->sqlind = new short(-1);	// NULL indicator
	}
}

void _XSQLDA::SetNull(int varnum)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::SetNull", "Variable index out of range.");

	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);
	if (! (var->sqltype & 1))
		throw ExceptionImpl("_XSQLDA::SetNull", "This parameter can't be null.");

	*var->sqlind = -1;	// Set the parameter to SQL NULL
}

void _XSQLDA::SetValue(int varnum, IITYPE ivType, void* value)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::SetValue", "Variable index out of range.");
	if (value == NULL)
		throw ExceptionImpl("_XSQLDA::SetValue", "Unexpected NULL pointer detected.");

	short len;
	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);
	switch (var->sqltype & ~1)
	{
		case SQL_TEXT :
			if (ivType == ivString)
			{
				len = (short)strlen((char*)value);
				if (len > var->sqllen) len = var->sqllen;
				strncpy(var->sqldata, (char*)value, len);
				while (len < var->sqllen) var->sqldata[len++] = ' ';
			}
			else if (ivType == ivDBKey)
			{
				IBPP::DBKey* key = (IBPP::DBKey*)value;
				key->GetKey(var->sqldata, var->sqllen);
			}
			else if (ivType == ivBool)
			{
				var->sqldata[0] = *(bool*)value ? 'T' : 'F';
				len = 1;
				while (len < var->sqllen) var->sqldata[len++] = ' ';
			}
			else throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
			break;

		case SQL_VARYING :
			if (ivType == ivString)
			{
				len = (short)strlen((char*)value);
				if (len > var->sqllen) len = var->sqllen;
				*(short*)var->sqldata = (short)len;
				strncpy(var->sqldata+2, (char*)value, len);
			}
			else if (ivType == ivBool)
			{
				*(short*)var->sqldata = (short)1;
				var->sqldata[2] = *(bool*)value ? 'T' : 'F';
			}
			else throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
			break;

		case SQL_SHORT :
			if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(short*)var->sqldata = *(bool*)value ? 1 : 0;
			}
			else if (ivType == ivShort)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(short*)var->sqldata = *(short*)value;
			}
			else if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				if (*(long*)value < minshort || *(long*)value > maxshort)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"Out of range numeric conversion !");
				*(short*)var->sqldata = (short)*(long*)value;
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				if (*(__int64*)value < minshort || *(__int64*)value > maxshort)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"Out of range numeric conversion !");
				*(short*)var->sqldata = (short)*(__int64*)value;
			}
			else if (ivType == ivFloat)
			{
				// This SQL_SHORT is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(short*)var->sqldata =
					(short)floor(*(float*)value * multiplier + 0.5);
			}
			else if (ivType == ivDouble)
			{
				// This SQL_SHORT is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(short*)var->sqldata =
					(short)floor(*(double*)value * multiplier + 0.5);
			}
			else throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
			break;

		case SQL_LONG :
			if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(long*)var->sqldata = *(bool*)value ? 1L : 0L;
			}
			else if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(long*)var->sqldata = *(long*)value;
			}
			else if (ivType == ivShort)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(long*)var->sqldata = *(short*)value;
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				if (*(__int64*)value < minlong || *(__int64*)value > maxlong)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"Out of range numeric conversion !");
				*(long*)var->sqldata = (long)*(__int64*)value;
			}
			else if (ivType == ivFloat)
			{
				// This SQL_LONG is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(long*)var->sqldata =
					(long)floor(*(float*)value * multiplier + 0.5);
			}
			else if (ivType == ivDouble)
			{
				// This SQL_LONG is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(long*)var->sqldata =
					(long)floor(*(double*)value * multiplier + 0.5);
			}
			else throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
			break;

		case SQL_INT64 :
			if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"Numeric(x,y) field : use IStatement::GetDouble()");
				*(__int64*)var->sqldata = *(bool*)value ? 1 : 0;
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"Numeric(x,y) field : use IStatement::GetDouble()");
				*(__int64*)var->sqldata = *(__int64*)value;
			}
			else if (ivType == ivShort)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(__int64*)var->sqldata = *(short*)value;
			}
			else if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(__int64*)var->sqldata = *(long*)value;
			}
			else if (ivType == ivFloat)
			{
				// This SQL_INT64 is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(__int64*)var->sqldata =
					(__int64)floor(*(float*)value * multiplier + 0.5);
			}
			else if (ivType == ivDouble)
			{
				// This SQL_INT64 is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(__int64*)var->sqldata =
					(__int64)floor(*(double*)value * multiplier + 0.5);
			}
			else throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
			break;

		case SQL_FLOAT :
			if (ivType != ivFloat || var->sqlscale != 0)
				throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
			*(float*)var->sqldata = *(float*)value;
			break;

		case SQL_DOUBLE :
			if (ivType != ivDouble) throw ExceptionImpl("_XSQLDA::SetValue",
										"Incompatible types.");
			if (var->sqlscale != 0)
			{
				// Round to scale of NUMERIC(x,y)
				double multiplier = dscales[-var->sqlscale];
				*(double*)var->sqldata =
					floor(*(double*)value * multiplier + 0.5) / multiplier;
			}
			else *(double*)var->sqldata = *(double*)value;
			break;

		case SQL_TIMESTAMP :
			/*
			if (GDS::gds()->_GDSVersion < 60)
			{
				// Version 5 Implementation : use an ISC_QUAD structure
				if (ivType != ivDate && ivType != ivTimestamp)
					throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
				IBPP::Date* date = (IBPP::Date*)value;
				//date->GetNative((ISC_QUAD*)var->sqldata);
				int year, month, day;
				date->GetDate(year, month, day);
				tm tms;
				memset(&tms, 0, sizeof(tms));
				tms.tm_year = year - 1900;
				tms.tm_mon = month - 1;
				tms.tm_mday = day;
				(*GDS::gds()->_encode_date)(&tms, (ISC_QUAD*)var->sqldata);
			}
			else
			{
			*/
				// Version 6 Implementation : use an ISC_TIMESTAMP structure
				if (ivType != ivTimestamp)
					throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
				{
					IBPP::Timestamp* timestamp = (IBPP::Timestamp*)value;
					//timestamp->GetNative((ISC_TIMESTAMP*)var->sqldata);
					int year, month, day;
					int hour, minute, second;
					timestamp->GetDate(year, month, day);
					timestamp->GetTime(hour, minute, second);
					tm tms;
					memset(&tms, 0, sizeof(tms));
					tms.tm_year = year - 1900;
					tms.tm_mon = month - 1;
					tms.tm_mday = day;
					tms.tm_hour = hour;
					tms.tm_min = minute;
					tms.tm_sec = second;
					(*GDS::gds()->_encode_timestamp)(&tms, (ISC_TIMESTAMP*)var->sqldata);
				}
			/*
			}
			*/
			break;

		case SQL_TYPE_DATE :
			if (ivType != ivDate) throw ExceptionImpl("_XSQLDA::SetValue",
										"Incompatible types.");
			{
				IBPP::Date* date = (IBPP::Date*)value;
				//date->GetNative((ISC_DATE*)var->sqldata);
				int year, month, day;
				date->GetDate(year, month, day);
				tm tms;
				memset(&tms, 0, sizeof(tms));
				tms.tm_year = year - 1900;
				tms.tm_mon = month - 1;
				tms.tm_mday = day;
				(*GDS::gds()->_encode_sql_date)(&tms, (ISC_DATE*)var->sqldata);
			}
			break;

		case SQL_TYPE_TIME :
			if (ivType != ivTime) throw ExceptionImpl("_XSQLDA::SetValue",
										"Incompatible types.");
			{
				IBPP::Time* time = (IBPP::Time*)value;
				//time->GetNative((ISC_TIME*)var->sqldata);
				int hour, minute, second;
				time->GetTime(hour, minute, second);
				tm tms;
				memset(&tms, 0, sizeof(tms));
				tms.tm_hour = hour;
				tms.tm_min = minute;
				tms.tm_sec = second;
				(*GDS::gds()->_encode_sql_time)(&tms, (ISC_TIME*)var->sqldata);
			}
			break;

		case SQL_BLOB :
			if (ivType != ivBlob) throw ExceptionImpl("_XSQLDA::SetValue",
										"Incompatible types.");
			{
				BlobImpl* blob = (BlobImpl*)value;
				blob->GetId((ISC_QUAD*)var->sqldata);
			}
			break;

		case SQL_ARRAY :
			if (ivType != ivArray) throw ExceptionImpl("_XSQLDA::SetValue",
										"Incompatible types.");
			{
				ArrayImpl* array = (ArrayImpl*)value;
				array->GetId((ISC_QUAD*)var->sqldata);
			}
			break;

		default : throw ExceptionImpl("_XSQLDA::SetValue",
						"The field uses an unsupported SQL type !");
	}

	if (var->sqltype & 1) *var->sqlind = 0;		// Remove the NULL flag
}

bool _XSQLDA::IsNull(int varnum)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::IsNull", "Variable index out of range.");

	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);
	return ((var->sqltype & 1) && *(var->sqlind) != 0) ? true : false;
}

void* _XSQLDA::GetValue(int varnum, IITYPE ivType, void* retvalue)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::GetValue", "Variable index out of range.");

	void* value;
	int len;
	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);

	// When there is no value (SQL NULL)
	if ((var->sqltype & 1) && *(var->sqlind) != 0) return NULL;

	switch (var->sqltype & ~1)
	{
		case SQL_TEXT :
			if (ivType == ivString)
			{
				var->sqldata[var->sqllen] = '\0';
				value = var->sqldata;
			}
			else if (ivType == ivDBKey)
			{
				IBPP::DBKey* key = (IBPP::DBKey*)retvalue;
				key->SetKey(var->sqldata, var->sqllen);
				value = retvalue;
			}
			else if (ivType == ivBool)
			{
				_Bools[varnum-1] = false;
				if (var->sqllen >= 1)
				{
					char c = var->sqldata[0];
					if (c == 't' || c == 'T' || c == 'y' || c == 'Y' ||	c == '1')
						_Bools[varnum-1] = true;
				}
				value = &_Bools[varnum-1];
			}
			else throw ExceptionImpl("_XSQLDA::GetValue", "Incompatible types.");
			break;

		case SQL_VARYING :
			if (ivType == ivString)
			{
				len = *(short*)var->sqldata;
				var->sqldata[len+2] = '\0';
				value = &var->sqldata[2];
			}
			else if (ivType == ivBool)
			{
				_Bools[varnum-1] = false;
				len = *(short*)var->sqldata;
				if (len >= 1)
				{
					char c = var->sqldata[2];
					if (c == 't' || c == 'T' || c == 'y' || c == 'Y' ||	c == '1')
						_Bools[varnum-1] = true;
				}
				value = &_Bools[varnum-1];
			}
			else throw ExceptionImpl("_XSQLDA::GetValue",
										"Incompatible types.");
			break;

		case SQL_SHORT :
			if (ivType == ivShort)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				value = var->sqldata;
			}
			else if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				if (*(short*)var->sqldata == 0)
					_Bools[varnum-1] = false;
				else _Bools[varnum-1] = true;
				value = &_Bools[varnum-1];
			}
			else if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				_Longs[varnum-1] = *(short*)var->sqldata;
				value = &_Longs[varnum-1];
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				_Larges[varnum-1] = *(short*)var->sqldata;
				value = &_Larges[varnum-1];
			}
			else if (ivType == ivFloat)
			{
				// This SQL_SHORT is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				_Floats[varnum-1] = (float)(*(short*)var->sqldata / divisor);

				value = &_Floats[varnum-1];
			}
			else if (ivType == ivDouble)
			{
				// This SQL_SHORT is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				_Numerics[varnum-1] = *(short*)var->sqldata / divisor;
				value = &_Numerics[varnum-1];
			}
			else throw ExceptionImpl("_XSQLDA::GetValue", "Incompatible types.");
			break;

		case SQL_LONG :
			if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				value = var->sqldata;
			}
			else if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				if (*(long*)var->sqldata == 0L)
					_Bools[varnum-1] = false;
				else _Bools[varnum-1] = true;
				value = &_Bools[varnum-1];
			}
			else if (ivType == ivShort)
			{
				long tmp;
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				tmp = *(long*)var->sqldata;
				if (tmp < minshort || tmp > maxshort)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"Out of range numeric conversion !");
				_Shorts[varnum-1] = (short)tmp;
				value = &_Shorts[varnum-1];
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				_Larges[varnum-1] = *(long*)var->sqldata;
				value = &_Larges[varnum-1];
			}
			else if (ivType == ivFloat)
			{
				// This SQL_LONG is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				_Floats[varnum-1] = (float)(*(long*)var->sqldata / divisor);
				value = &_Floats[varnum-1];
			}
			else if (ivType == ivDouble)
			{
				// This SQL_LONG is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				_Numerics[varnum-1] = *(long*)var->sqldata / divisor;
				value = &_Numerics[varnum-1];
			}
			else throw ExceptionImpl("_XSQLDA::GetValue", "Incompatible types.");
			break;

		case SQL_INT64 :
			if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				value = var->sqldata;
			}
			else if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				if (*(__int64*)var->sqldata == 0)
					_Bools[varnum-1] = false;
				else _Bools[varnum-1] = true;
				value = &_Bools[varnum-1];
			}
			else if (ivType == ivShort)
			{
				__int64 tmp;
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				tmp = *(__int64*)var->sqldata;
				if (tmp < minshort || tmp > maxshort)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"Out of range numeric conversion !");
				_Shorts[varnum-1] = (short)tmp;
				value = &_Shorts[varnum-1];
			}
			else if (ivType == ivLong)
			{
				__int64 tmp;
				if (var->sqlscale != 0)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				tmp = *(__int64*)var->sqldata;
				if (tmp < minlong || tmp > maxlong)
					throw ExceptionImpl("_XSQLDA::GetValue",
						"Out of range numeric conversion !");
				_Longs[varnum-1] = (short)tmp;
				value = &_Longs[varnum-1];
			}
			else if (ivType == ivFloat)
			{
				// This SQL_INT64 is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				_Floats[varnum-1] = (float)(*(__int64*)var->sqldata / divisor);
				value = &_Floats[varnum-1];
			}
			else if (ivType == ivDouble)
			{
				// This SQL_INT64 is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				_Numerics[varnum-1] = *(__int64*)var->sqldata / divisor;
				value = &_Numerics[varnum-1];
			}
			else throw ExceptionImpl("_XSQLDA::GetValue", "Incompatible types.");
			break;

		case SQL_FLOAT :
			if (ivType != ivFloat) throw ExceptionImpl("_XSQLDA::GetValue",
										"Incompatible types.");
			value = var->sqldata;
			break;

		case SQL_DOUBLE :
			if (ivType != ivDouble) throw ExceptionImpl("_XSQLDA::GetValue",
										"Incompatible types.");
			if (var->sqlscale != 0)
			{
				// Round to scale y of NUMERIC(x,y)
				double multiplier = dscales[-var->sqlscale];
				_Numerics[varnum-1] =
					floor(*(double*)var->sqldata * multiplier + 0.5) / multiplier;
				value = &_Numerics[varnum-1];
			}
			else value = var->sqldata;
			break;

		case SQL_TIMESTAMP :
			/*
			if (GDS::gds()->_GDSVersion < 60)
			{
				// Version 5 implementation
				if (ivType != ivDate && ivType != ivTimestamp)
					throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
				DateImpl* date = (DateImpl*)retvalue;
				date->SetNative((ISC_QUAD*)var->sqldata);
				value = retvalue;
			}
			else
			{
			*/
				// Version 6 implementation
				if (ivType != ivTimestamp)
					throw ExceptionImpl("_XSQLDA::SetValue", "Incompatible types.");
				{
					IBPP::Timestamp* timestamp = (IBPP::Timestamp*)retvalue;
					//timestamp->SetNative((ISC_TIMESTAMP*)var->sqldata);
					tm tms;
					(*GDS::gds()->_decode_timestamp)((ISC_TIMESTAMP*)var->sqldata, &tms);
					timestamp->SetDate(tms.tm_year + 1900, tms.tm_mon + 1, tms.tm_mday);
					timestamp->SetTime(tms.tm_hour, tms.tm_min, tms.tm_sec);
					value = retvalue;
				}
			/*
			}
			*/
			break;

		case SQL_TYPE_DATE :
			if (ivType != ivDate) throw ExceptionImpl("_XSQLDA::SetValue",
										"Incompatible types.");
			{
				IBPP::Date* date = (IBPP::Date*)retvalue;
				//date->SetNative((ISC_DATE*)var->sqldata);
				tm tms;
				(*GDS::gds()->_decode_sql_date)((ISC_DATE*)var->sqldata, &tms);
				date->SetDate(tms.tm_year + 1900, tms.tm_mon + 1, tms.tm_mday);
				value = retvalue;
			}
			break;

		case SQL_TYPE_TIME :
			if (ivType != ivTime) throw ExceptionImpl("_XSQLDA::SetValue",
										"Incompatible types.");
			{
				IBPP::Time* time = (IBPP::Time*)retvalue;
				//time->SetNative((ISC_TIME*)var->sqldata);
				tm tms;
				(*GDS::gds()->_decode_sql_time)((ISC_TIME*)var->sqldata, &tms);
				time->SetTime(tms.tm_hour, tms.tm_min, tms.tm_sec);
				value = retvalue;
			}
			break;

		case SQL_BLOB :
			if (ivType != ivBlob) throw ExceptionImpl("_XSQLDA::GetValue",
										"Incompatible types.");
			{
				BlobImpl* blob = (BlobImpl*)retvalue;
				blob->SetId((ISC_QUAD*)var->sqldata);
				value = retvalue;
			}
			break;

		case SQL_ARRAY :
			if (ivType != ivArray) throw ExceptionImpl("_XSQLDA::GetValue",
										"Incompatible types.");
			{
				ArrayImpl* array = (ArrayImpl*)retvalue;
				array->SetId((ISC_QUAD*)var->sqldata);
				value = retvalue;
			}
			break;

		default : throw ExceptionImpl("_XSQLDA::GetValue",
						"Found an unknown sqltype !");
	}

	return value;
}

int _XSQLDA::GetVarNum(const char* name)
{
	XSQLVAR* var;
	const int sqlnamelen = sizeof(var->sqlname);
	char Uname[sqlnamelen+1];		// Max size of sqlname + '\0'

	if (name == NULL)
		throw ExceptionImpl("_XSQLDA::GetVarNum", "Unexpected NULL pointer detected.");

	// Local upper case copy of the column name
	strncpy(Uname, name, sqlnamelen);
	Uname[sqlnamelen] = '\0';
	int Len = strlen(strupr(Uname));

	// Loop through the columns of the descriptor
	for (int i = 0; i < _DescrArea->sqld; i++)
	{
		var = &(_DescrArea->sqlvar[i]);
		if (var->sqlname_length != (short)Len) continue;
		if (strncmp(Uname, var->sqlname, Len) == 0) return i+1;
	}

	return 0;
}

int _XSQLDA::GetVarNumAlias(const char* alias)
{
	XSQLVAR* var;
	const int aliasnamelen = sizeof(var->aliasname);
	char Ualias[aliasnamelen+1];		// Max size of aliasname + '\0'

	if (alias == NULL)
		throw ExceptionImpl("_XSQLDA::GetVarNumAlias", "Unexpected NULL pointer detected.");

	// Local upper case copy of the column name
	strncpy(Ualias, alias, aliasnamelen);
	Ualias[aliasnamelen] = '\0';
	int Len = strlen(strupr(Ualias));

	// Loop through the columns of the descriptor
	for (int i = 0; i < _DescrArea->sqld; i++)
	{
		var = &(_DescrArea->sqlvar[i]);
		if (var->aliasname_length != (short)Len) continue;
		if (strncmp(Ualias, var->aliasname, Len) == 0) return i+1;
	}

	return 0;
}
/*
GetName, GetAlias, GetTable : all these 3 have a mistake.
Ideally, the strings should be stored elsewhere (like _Numerics and so on) to
take into account the final '\0' which needs to be added. For now, we insert
the '\0' in the original data, which will cut the 32th character. Not terribly
bad, but should be cleanly rewritten.
*/

const char* _XSQLDA::GetName(int varnum)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::GetName", "Variable index out of range.");

	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);
	if (var->sqlname_length >= 31) var->sqlname_length = 31;
	var->sqlname[var->sqlname_length] = '\0';
	return var->sqlname;
}

const char* _XSQLDA::GetAlias(int varnum)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::GetAlias", "Variable index out of range.");

	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);
	if (var->aliasname_length >= 31) var->aliasname_length = 31;
	var->aliasname[var->aliasname_length] = '\0';
	return var->aliasname;
}

const char* _XSQLDA::GetTable(int varnum)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::GetTable", "Variable index out of range.");

	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);
	if (var->relname_length >= 31) var->relname_length = 31;
	var->relname[var->relname_length] = '\0';
	return var->relname;
}

IBPP::SDT _XSQLDA::GetType(int varnum)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::GetType", "Variable index out of range.");

	IBPP::SDT value;
	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);

	switch (var->sqltype & ~1)
	{
		case SQL_TEXT :      value = IBPP::sdString;    break;
		case SQL_VARYING :   value = IBPP::sdString;    break;
		case SQL_SHORT :     value = IBPP::sdSmallint;  break;
		case SQL_LONG :      value = IBPP::sdInteger;   break;
		case SQL_INT64 :     value = IBPP::sdLargeint;  break;
		case SQL_FLOAT :     value = IBPP::sdFloat;     break;
		case SQL_DOUBLE :    value = IBPP::sdDouble;    break;
		case SQL_TIMESTAMP : value = IBPP::sdTimestamp; break;
		case SQL_TYPE_DATE : value = IBPP::sdDate;      break;
		case SQL_TYPE_TIME : value = IBPP::sdTime;      break;
		case SQL_BLOB :      value = IBPP::sdBlob;      break;
		case SQL_ARRAY :     value = IBPP::sdArray;     break;
		default : throw ExceptionImpl("_XSQLDA::GetType",
						"Found an unknown sqltype !");
	}

	return value;
}

int _XSQLDA::GetSize(int varnum)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::GetSize", "Variable index out of range.");

	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);
    return var->sqllen;
}

int _XSQLDA::GetScale(int varnum)
{
	if (varnum < 1 || varnum > _DescrArea->sqld)
		throw ExceptionImpl("_XSQLDA::GetScale", "Variable index out of range.");

	XSQLVAR* var = &(_DescrArea->sqlvar[varnum-1]);
    return -var->sqlscale;
}

_XSQLDA::_XSQLDA()
{
	_DescrArea = NULL;
	_Numerics = NULL;
	_Floats = NULL;
	_Larges = NULL;
	_Longs = NULL;
	_Shorts = NULL;
	_Bools = NULL;
	Resize(1);
}

_XSQLDA::_XSQLDA(const _XSQLDA& copied)
{
	int n = copied._DescrArea->sqln;
	int size = XSQLDA_LENGTH(n);
	int i;

	// Initial brute copy
    _DescrArea = (XSQLDA*) new char[size];
	memcpy(_DescrArea, copied._DescrArea, size);

	// Pointers init, real data copy
	_Numerics = new double[n];
	memcpy(_Numerics, copied._Numerics, sizeof(double)*n);
	_Floats = new float[n];
	memcpy(_Floats, copied._Floats, sizeof(float)*n);
	_Larges = new __int64[n];
	memcpy(_Larges, copied._Larges, sizeof(__int64)*n);
	_Longs = new long[n];
	memcpy(_Longs, copied._Longs, sizeof(long)*n);
	_Shorts = new short[n];
	memcpy(_Shorts, copied._Shorts, sizeof(short)*n);
	_Bools = new bool[n];
	memcpy(_Bools, copied._Bools, sizeof(bool)*n);

	// Copy of the columns data
	for (i = 0; i < _DescrArea->sqld; i++)
	{
		XSQLVAR* var = &(_DescrArea->sqlvar[i]);
		XSQLVAR* org = &(copied._DescrArea->sqlvar[i]);
		switch (var->sqltype & ~1)
		{
			case SQL_ARRAY :
			case SQL_BLOB :		var->sqldata = (char*) new ISC_QUAD;
								memcpy(var->sqldata, org->sqldata, sizeof(ISC_QUAD));
								break;
			case SQL_TIMESTAMP :var->sqldata = (char*) new ISC_TIMESTAMP;
								memcpy(var->sqldata, org->sqldata, sizeof(ISC_TIMESTAMP));
								break;
			case SQL_TYPE_TIME :var->sqldata = (char*) new ISC_TIME;
								memcpy(var->sqldata, org->sqldata, sizeof(ISC_TIME));
								break;
			case SQL_TYPE_DATE :var->sqldata = (char*) new ISC_DATE;
								memcpy(var->sqldata, org->sqldata, sizeof(ISC_DATE));
								break;
			case SQL_TEXT :		var->sqldata = new char[var->sqllen+1];
								memcpy(var->sqldata, org->sqldata, var->sqllen+1);
								break;
			case SQL_VARYING :	var->sqldata = new char[var->sqllen+3];
								memcpy(var->sqldata, org->sqldata, var->sqllen+3);
								break;
			case SQL_SHORT :	var->sqldata = (char*) new short(*(short*)org->sqldata); break;
			case SQL_LONG :		var->sqldata = (char*) new long(*(long*)org->sqldata); break;
			case SQL_INT64 :	var->sqldata = (char*) new __int64(*(__int64*)org->sqldata); break;
			case SQL_FLOAT : 	var->sqldata = (char*) new float(*(float*)org->sqldata); break;
			case SQL_DOUBLE :	var->sqldata = (char*) new double(*(double*)org->sqldata); break;
			default : throw ExceptionImpl("_XSQLDA::Ctor",
						"Found an unknown sqltype !");
		}
		if (var->sqltype & 1) var->sqlind = new short(*org->sqlind);	// NULL indicator
	}
}

_XSQLDA::_XSQLDA(short n)
{
	_DescrArea = NULL;
	_Numerics = NULL;
	_Floats = NULL;
	_Larges = NULL;
	_Longs = NULL;
	_Shorts = NULL;
	_Bools = NULL;
	Resize(n);
}

_XSQLDA::~_XSQLDA()
{
	Free();
}

/*
Documentation purpose.
Just a reminder of the Standard C Library "tm" structure.
This is used in methods SetValue and GetValue when dealing with dates/times.

struct tm {
        int tm_sec;     // seconds after the minute - [0,59]
        int tm_min;     // minutes after the hour - [0,59]
        int tm_hour;    // hours since midnight - [0,23]
        int tm_mday;    // day of the month - [1,31]
        int tm_mon;     // months since January - [0,11]
        int tm_year;    // years since 1900 - [0, ?]
        int tm_wday;    // days since Sunday - [0,6]
        int tm_yday;    // days since January 1 - [0,365]
        int tm_isdst;   // daylight savings time flag
        };
*/

//
//	EOF
//
