///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: array.cpp,v 1.6 2001/11/03 21:01:43 epocman Exp $
//	Subject : IBPP, Array class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

//	(((((((( OBJECT INTERFACE IMPLEMENTATION ))))))))

void ArrayImpl::AttachDatabase(IBPP::Database* database)
{
	if (database == NULL) throw ExceptionImpl("Array::AttachDatabase",
			"Can't attach a NULL Database object.");

	if (_Database != NULL) _Database->DetachArray(this);
	_Database = dynamic_cast<DatabaseImpl*>(database);
	_Database->AttachArray(this);
}

void ArrayImpl::AttachTransaction(IBPP::Transaction* transaction)
{
	if (transaction == NULL) throw ExceptionImpl("Array::AttachTransaction",
			"Can't attach a NULL Transaction object.");

	if (_Transaction != NULL) _Transaction->DetachArray(this);
	_Transaction = dynamic_cast<TransactionImpl*>(transaction);
	_Transaction->AttachArray(this);
}

void ArrayImpl::DetachDatabase(void)
{
	if (_Database == NULL) throw ExceptionImpl("Array::DetachDatabase",
			"No Database was attached.");

	_Database->DetachArray(this);
	_Database = NULL;
}

void ArrayImpl::DetachTransaction(void)
{
	if (_Transaction == NULL) throw ExceptionImpl("Array::DetachTransaction",
			"No Transaction was attached.");

	_Transaction->DetachArray(this);
	_Transaction = NULL;
}

IBPP::Database* ArrayImpl::GetDatabase(void)
{
	if (_Database == NULL) throw ExceptionImpl("Array::GetDatabase",
			"No Database is attached.");
	return _Database;
}

IBPP::Transaction* ArrayImpl::GetTransaction(void)
{
	if (_Transaction == NULL) throw ExceptionImpl("Array::GetTransaction",
			"No Transaction is attached.");
	return _Transaction;
}

/*
void ArrayImpl::Open(void)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Array::Open", "Array already opened.");
	if (_Database == NULL)
		throw ExceptionImpl("Array::Open", "No Database is attached.");
	if (_Transaction == NULL)
		throw ExceptionImpl("Array::Open", "No Transaction is attached.");
	if (! _IdAssigned)
		throw ExceptionImpl("Array::Open", "Array Id is not assigned.");

	_IBS status;
	(*GDS::gds()->_open_blob2)(status.Self(), _Database->GetHandlePtr(),
		_Transaction->GetHandlePtr(), &_Handle, &_Id, 0, NULL);
	if (status.Errors())
		throw ExceptionImpl(&status, "Array::Open", "isc_open_blob2 failed.");
	_WriteMode = false;
}

void ArrayImpl::Create(void)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Array::Create", "Array already opened.");
	if (_Database == NULL)
		throw ExceptionImpl("Array::Create", "No Database is attached.");
	if (_Transaction == NULL)
		throw ExceptionImpl("Array::Create", "No Transaction is attached.");

	_IBS status;
	(*GDS::gds()->_create_blob)(status.Self(), _Database->GetHandlePtr(),
		_Transaction->GetHandlePtr(), &_Handle, &_Id);
	if (status.Errors())
		throw ExceptionImpl(&status, "Array::Create",
			"isc_create_blob failed.");
	_IdAssigned = true;
	_WriteMode = true;
}

void ArrayImpl::Close(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Array::Close", "The Array is not opened");

	_IBS status;
	(*GDS::gds()->_close_blob)(status.Self(), &_Handle);
	if (status.Errors())
		throw ExceptionImpl(&status, "Array::Close", "isc_close_blob failed.");
	_Handle = NULL;
}

void ArrayImpl::Cancel(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Array::Cancel", "The Array is not opened");
	if (! _WriteMode)
		throw ExceptionImpl("Array::Cancel", "Can't cancel a Array opened for read");

	_IBS status;
	(*GDS::gds()->_cancel_blob)(status.Self(), &_Handle);
	if (status.Errors())
		throw ExceptionImpl(&status, "Array::Cancel", "isc_cancel_blob failed.");
	_Handle = NULL;
	_IdAssigned = false;
}

int ArrayImpl::Read(void* buffer, int size)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Array::Read", "The Array is not opened");
	if (_WriteMode)
		throw ExceptionImpl("Array::Read", "Can't read from Array opened for write");
	if (size < 1 || size > (64*1024-1))
		throw ExceptionImpl("Array::Read", "Invalid segment size (max 64Kb-1)");

	_IBS status;
	unsigned short bytesread;
	int result = (*GDS::gds()->_get_segment)(status.Self(), &_Handle, &bytesread,
					(unsigned short)size, (char*)buffer);
	if (result == isc_segstr_eof) return 0;	// Fin du blob
	if (result != isc_segment && status.Errors())
		throw ExceptionImpl(&status, "Array::Read", "isc_get_segment failed.");
	return (int)bytesread;
}

void ArrayImpl::Write(void* buffer, int size)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Array::Write", "The Array is not opened");
	if (! _WriteMode)
		throw ExceptionImpl("Array::Write", "Can't write to Array opened for read");
	if (size < 1 || size > (64*1024-1))
		throw ExceptionImpl("Array::Write", "Invalid segment size (max 64Kb-1)");

	_IBS status;
	(*GDS::gds()->_put_segment)(status.Self(), &_Handle,
		(unsigned short)size, (char*)buffer);
	if (status.Errors())
		throw ExceptionImpl(&status, "Array::Write", "isc_put_segment failed.");
}

void ArrayImpl::GetInfo(int* Size, int* Largest, int* Segments)
{
	char items[] = {isc_info_blob_total_length,
					isc_info_blob_max_segment,
					isc_info_blob_num_segments};

	if (_Handle == NULL)
		throw ExceptionImpl("Array::GetInfo", "The Array is not opened");

	_IBS status;
	_RB result(100);
	(*GDS::gds()->_blob_info)(status.Self(), &_Handle, sizeof(items), items,
		result.Size(), result.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Array::GetInfo", "isc_blob_info failed.");

	if (Size != NULL) *Size = result.GetValue(isc_info_blob_total_length);
	if (Largest != NULL) *Largest = result.GetValue(isc_info_blob_max_segment);
	if (Segments != NULL) *Segments = result.GetValue(isc_info_blob_num_segments);
}
*/

void ArrayImpl::Release(IBPP::Array*& Self)
{
	if (this != dynamic_cast<ArrayImpl*>(Self))
		throw ExceptionImpl("Array::Release", "Invalid Release()");
	delete this;
	Self = NULL;
}

//	(((((((( OBJECT INTERNAL METHODS ))))))))

void ArrayImpl::Init(void)
{
	_IdAssigned = false;
	_Database = NULL;
	_Transaction = NULL;
	/*
	_Handle = NULL;
	_WriteMode = false;
	*/
}

void ArrayImpl::SetId(ISC_QUAD* quad)
{
	/*
	if (_Handle != NULL)
		throw ExceptionImpl("ArrayImpl::SetId", "Can't set Id on an opened ArrayImpl.");
	*/
	if (quad == NULL)
		throw ExceptionImpl("ArrayImpl::SetId", "NULL Id reference detected.");

	memcpy(&_Id, quad, sizeof(_Id));
	_IdAssigned = true;
}

void ArrayImpl::GetId(ISC_QUAD* quad)
{
	/*
	if (_Handle != NULL)
		throw ExceptionImpl("ArrayImpl::GetId", "Can't get Id on an opened ArrayImpl.");
	if (! _WriteMode)
		throw ExceptionImpl("ArrayImpl::GetId", "Can only get Id of a newly created Array.");
	*/
	if (quad == NULL)
		throw ExceptionImpl("ArrayImpl::GetId", "NULL Id reference detected.");

	memcpy(quad, &_Id, sizeof(_Id));
}

ArrayImpl::ArrayImpl(DatabaseImpl* database, TransactionImpl* transaction)
{
	Init();
	AttachDatabase(database);
	if (transaction != NULL) AttachTransaction(transaction);
}

ArrayImpl::~ArrayImpl()
{
	/*
	if (_Handle != NULL)
	{
		if (_WriteMode) Cancel();
		else try {Close();} catch (IBPP::Exception&) {}
	}
	*/
	if (_Transaction != NULL) _Transaction->DetachArray(this);
	if (_Database != NULL) _Database->DetachArray(this);
}

//
//	EOF
//
