///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: database.cpp,v 1.12 2001/11/10 12:52:58 epocman Exp $
//	Subject : IBPP, Database class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

//	(((((((( OBJECT INTERFACE IMPLEMENTATION ))))))))

void DatabaseImpl::SetServerName(const char* newName)
{
	if (_ServerName != NULL) delete [] _ServerName;
	if (newName == NULL) _ServerName = NULL;
	else
	{
		_ServerName = new char[strlen(newName)+1];
		strcpy(_ServerName, newName);
	}
}

void DatabaseImpl::SetDatabaseName(const char* newName)
{
	if (_DatabaseName != NULL) delete [] _DatabaseName;
	if (newName == NULL) _DatabaseName = NULL;
	else
	{
		_DatabaseName = new char[strlen(newName)+1];
		strcpy(_DatabaseName, newName);
	}
}

void DatabaseImpl::SetUsername(const char* newName)
{
	if (_UserName != NULL) delete [] _UserName;
	if (newName == NULL) _UserName = NULL;
	else
	{
		_UserName = new char[strlen(newName)+1];
		strcpy(_UserName, newName);
	}
}

void DatabaseImpl::SetUserPassword(const char* newPassword)
{
	if (_UserPassword != NULL) delete [] _UserPassword;
	if (newPassword == NULL) _UserPassword = NULL;
	else
	{
		_UserPassword = new char[strlen(newPassword)+1];
		strcpy(_UserPassword, newPassword);
	}
}

void DatabaseImpl::SetRoleName(const char* newRole)
{
	if (_RoleName != NULL) delete [] _RoleName;
	if (newRole == NULL) _RoleName = NULL;
	else
	{
		_RoleName = new char[strlen(newRole)+1];
		strcpy(_RoleName, newRole);
	}
}

void DatabaseImpl::SetCreateParams(const char* newParams)
{
	if (_CreateParams != NULL) delete [] _CreateParams;
	if (newParams == NULL) _CreateParams = NULL;
	else
	{
		_CreateParams = new char[strlen(newParams)+1];
		strcpy(_CreateParams, newParams);
	}
}

const char* DatabaseImpl::GetServerName(void)
{
	return _ServerName;
}

const char* DatabaseImpl::GetDatabaseName(void)
{
	return _DatabaseName;
}

const char* DatabaseImpl::GetUsername(void)
{
	return _UserName;
}

const char* DatabaseImpl::GetUserPassword(void)
{
	return _UserPassword;
}

const char* DatabaseImpl::GetRoleName(void)
{
	return _RoleName;
}

const char* DatabaseImpl::GetCreateParams(void)
{
	return _CreateParams;
}

void DatabaseImpl::AttachTransaction(IBPP::Transaction* tr)
{
	if (tr == NULL)
		throw ExceptionImpl("Database::AttachTransaction",
					"Transaction object is null.");

	// Alloue un nouveau tableau
	TransactionImpl** newTransactions = new TransactionImpl* [_TrCount+1];
	// Recopie l'ancien, s'il existait
	if (_TrCount > 0)
		memcpy((BYTE*)newTransactions, (BYTE*)_Transactions,
			_TrCount*sizeof(TransactionImpl*));
	// Ajoute le nouvel lment
	newTransactions[_TrCount] = dynamic_cast<TransactionImpl*>(tr);
	// Dtruit l'ancien tableau
	if (_Transactions != NULL) delete [] _Transactions;
	// Affecte le nouveau tableau
	_Transactions = newTransactions;
	// Incrmente le nombre de transactions associes
	_TrCount++;
		
	return;
}

void DatabaseImpl::DetachTransaction(IBPP::Transaction* tr)
{
	if (tr == NULL)
		throw ExceptionImpl("Database::DetachTransaction",
				"ITransaction object is null.");

	// Recherche cette transaction
	for (int i = 0; i < _TrCount; i++)
	{
		ASSERTION(_Transactions != NULL)
		ASSERTION(_Transactions[i] != NULL)
		if (_Transactions[i] == dynamic_cast<TransactionImpl*>(tr))
		{
			// La transaction dont il faut se dtacher est identifie.
			// Eliminons-la du tableau par dcalage des autres lments.
			// Il importe peu qu'un lment allou devienne inutilis. Si on
			// attache DatabaseImpl  une autre transaction, un nouveau tableau
			// sera de toute faon allou.
			for (int j = i; j < _TrCount-1; j++)
				_Transactions[j] = _Transactions[j+1];
			_TrCount--;
			ASSERTION(_TrCount >= 0)
			break;
		}
	}
}

void DatabaseImpl::AttachStatement(IBPP::Statement* st)
{
	if (st == NULL)
		throw ExceptionImpl("Database::AttachStatement",
					"Can't attach a NULL Statement object.");

	// Alloue un nouveau tableau
	StatementImpl** newStatements = new StatementImpl* [_StCount+1];
	// Recopie l'ancien, s'il existait
	if (_StCount > 0)
		memcpy((BYTE*)newStatements, (BYTE*)_Statements,
			_StCount*sizeof(StatementImpl*));
	// Ajoute le nouvel lment
	newStatements[_StCount] = dynamic_cast<StatementImpl*>(st);
	// Dtruit l'ancien tableau
	if (_Statements != NULL) delete [] _Statements;
	// Affecte le nouveau tableau
	_Statements = newStatements;
	// Incrmente le nombre de transactions associes
	_StCount++;
		
	return;
}

void DatabaseImpl::DetachStatement(IBPP::Statement* st)
{
	if (st == NULL)
		throw ExceptionImpl("Database::DetachStatement",
				"Can't detach a NULL Statement object.");

	// Recherche cette transaction
	for (int i = 0; i < _StCount; i++)
	{
		ASSERTION(_Statements != NULL)
		ASSERTION(_Statements[i] != NULL)
		if (_Statements[i] == dynamic_cast<IBPP::Statement*>(st))
		{
			// Le statement dont il faut se dtacher est identifi.
			// Eliminons-la du tableau par dcalage des autres lments.
			// Il importe peu qu'un lment allou devienne inutilis. Si on
			// attache DatabaseImpl  un autre statement, un nouveau tableau
			// sera de toute faon allou.
			for (int j = i; j < _StCount-1; j++)
				_Statements[j] = _Statements[j+1];
			_StCount--;
			ASSERTION(_StCount >= 0)
			break;
		}
	}
}

void DatabaseImpl::AttachBlob(IBPP::Blob* bb)
{
	if (bb == NULL)
		throw ExceptionImpl("Database::AttachBlob",
					"Can't attach a NULL Blob object.");

	// Alloue un nouveau tableau
	BlobImpl** newBlobs = new BlobImpl* [_BbCount+1];
	// Recopie l'ancien, s'il existait
	if (_BbCount > 0)
		memcpy((BYTE*)newBlobs, (BYTE*)_Blobs,
			_BbCount*sizeof(BlobImpl*));
	// Ajoute le nouvel lment
	newBlobs[_BbCount] = dynamic_cast<BlobImpl*>(bb);
	// Dtruit l'ancien tableau
	if (_Blobs != NULL) delete [] _Blobs;
	// Affecte le nouveau tableau
	_Blobs = newBlobs;
	// Incrmente le nombre de blobs attachs
	_BbCount++;
		
	return;
}

void DatabaseImpl::DetachBlob(IBPP::Blob* bb)
{
	if (bb == NULL)
		throw ExceptionImpl("Database::DetachBlob",
				"Can't detach a NULL Blob object.");

	// Recherche ce Blob
	for (int i = 0; i < _BbCount; i++)
	{
		ASSERTION(_Blobs != NULL)
		ASSERTION(_Blobs[i] != NULL)
		if (_Blobs[i] == dynamic_cast<IBPP::Blob*>(bb))
		{
			// Le blob dont il faut se dtacher est identifi.
			// Eliminons-le du tableau par dcalage des autres lments.
			// Il importe peu qu'un lment allou devienne inutilis. Si on
			// attache DatabaseImpl  un autre blob, un nouveau tableau
			// sera de toute faon allou.
			for (int j = i; j < _BbCount-1; j++)
				_Blobs[j] = _Blobs[j+1];
			_BbCount--;
			ASSERTION(_BbCount >= 0)
			break;
		}
	}
}

void DatabaseImpl::AttachArray(IBPP::Array* ar)
{
	if (ar == NULL)
		throw ExceptionImpl("Database::AttachArray",
					"Can't attach a NULL Array object.");

	// Alloue un nouveau tableau
	ArrayImpl** newArrays = new ArrayImpl* [_ArCount+1];
	// Recopie l'ancien, s'il existait
	if (_ArCount > 0)
		memcpy((BYTE*)newArrays, (BYTE*)_Arrays,
			_ArCount*sizeof(ArrayImpl*));
	// Ajoute le nouvel lment
	newArrays[_ArCount] = dynamic_cast<ArrayImpl*>(ar);
	// Dtruit l'ancien tableau
	if (_Arrays != NULL) delete [] _Arrays;
	// Affecte le nouveau tableau
	_Arrays = newArrays;
	// Incrmente le nombre de arrays attachs
	_ArCount++;
		
	return;
}

void DatabaseImpl::DetachArray(IBPP::Array* ar)
{
	if (ar == NULL)
		throw ExceptionImpl("Database::DetachArray",
				"Can't detach a NULL Array object.");

	// Recherche cet Array
	for (int i = 0; i < _ArCount; i++)
	{
		ASSERTION(_Arrays != NULL)
		ASSERTION(_Arrays[i] != NULL)
		if (_Arrays[i] == dynamic_cast<IBPP::Array*>(ar))
		{
			// Le Array dont il faut se dtacher est identifi.
			// Eliminons-le du tableau par dcalage des autres lments.
			// Il importe peu qu'un lment allou devienne inutilis. Si on
			// attache DatabaseImpl  un autre array, un nouveau tableau
			// sera de toute faon allou.
			for (int j = i; j < _ArCount-1; j++)
				_Arrays[j] = _Arrays[j+1];
			_ArCount--;
			ASSERTION(_ArCount >= 0)
			break;
		}
	}
}

void DatabaseImpl::Create(void)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Database::Create", "Database is already connected.");
	if (_DatabaseName == NULL)
		throw ExceptionImpl("Database::Create", "Unspecified database name.");
	if (_UserName == NULL)
		throw ExceptionImpl("Database::Create", "Unspecified user name.");
	if (_UserPassword == NULL)
		throw ExceptionImpl("Database::Create", "Unspecified user password.");

    // Create a new database
    _IBS status;
	char connect[256];
    char create[1024];
    void* tr_handle = NULL;

    // Build the SQL Create Statement
	if (_ServerName != NULL)
	{
		strcpy(connect, _ServerName);
		strcat(connect, ":");
	}
	else connect[0] = '\0';
	strcat(connect, _DatabaseName);

	sprintf(create, "CREATE DATABASE \"%s\" USER \"%s\" PASSWORD \"%s\" ",
		connect, _UserName, _UserPassword);
	if (_CreateParams != NULL)
	{
		strcat(create, " ");
		strcat(create, _CreateParams);
	}

    (*GDS::gds()->_dsql_execute_immediate)(status.Self(), &_Handle, &tr_handle,
    	0, create, 3, NULL);
    if (status.Errors())
		throw ExceptionImpl(&status, "Database::Create", "isc_dsql_execute_immediate failed");
	Disconnect();
}

void DatabaseImpl::Connect(void)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Database::Connect", "Database is already connected.");
	if (_DatabaseName == NULL)
		throw ExceptionImpl("Database::Connect", "Unspecified database name.");
	if (_UserName == NULL)
		throw ExceptionImpl("Database::Connect", "Unspecified user name.");
	if (_UserPassword == NULL)
		throw ExceptionImpl("Database::Connect", "Unspecified user password.");

	// Attach to the database
    _IBS status;
    _DPB dpb;
	char connect[1024];

    // Build a DPB based on the properties
    dpb.Insert(isc_dpb_user_name, _UserName);
    dpb.Insert(isc_dpb_password, _UserPassword);
    dpb.Insert(isc_dpb_force_write, (short)0);
    if (_RoleName != NULL) dpb.Insert(isc_dpb_sql_role_name, _RoleName);

	if (_ServerName != NULL)
	{
		strcpy(connect, _ServerName);
		strcat(connect, ":");
	}
	else connect[0] = '\0';
	strcat(connect, _DatabaseName);

    (*GDS::gds()->_attach_database)(status.Self(), (short)strlen(connect), connect,
    	&_Handle, dpb.Size(), dpb.Self());
    if (status.Errors())
    {
        _Handle = NULL;     // Should be, but better be sure...
		throw ExceptionImpl(&status, "Database::Connect", "isc_attach_database failed");
    }

	// Now, get ODS version information and dialect.
	// If ODS major is 9, this an InterBase 5.x Server (no dialect).
	// If ODS major is 10, this is an InterBase 6.x. Server.
	// Getting the ODS also gives us the IBSERVER version, because IB 6 cannot
	// access nor upgrade previous ODS to ODS 10.

	char items[] = {isc_info_ods_version,
					isc_info_db_SQL_dialect,
					isc_info_end};
	_RB result(100);

	status.Reset();
	(*GDS::gds()->_database_info)(status.Self(), &_Handle, sizeof(items), items,
		result.Size(), result.Self());
	if (status.Errors())
	{
		status.Reset();
	    (*GDS::gds()->_detach_database)(status.Self(), &_Handle);
        _Handle = NULL;     // Should be, but better be sure...
		throw ExceptionImpl(&status, "Database::Connect", "isc_database_info failed");
	}

	int ODS = result.GetValue(isc_info_ods_version);
	//if (ODS == 9) _Dialect = 1;			// InterBase 5, 5.1, 5.5, 5.6
	/*else*/ if (ODS == 10)
	{
		_Dialect = result.GetValue(isc_info_db_SQL_dialect);
		if (_Dialect < 3)
		{
			status.Reset();
		    (*GDS::gds()->_detach_database)(status.Self(), &_Handle);
	        _Handle = NULL;     // Should be, but better be sure...
			throw ExceptionImpl("Database::Connect", "Dialect 3 required");
		}
	}
	else
	{
		status.Reset();
	    (*GDS::gds()->_detach_database)(status.Self(), &_Handle);
        _Handle = NULL;     // Should be, but better be sure...

		throw ExceptionImpl("Database::Connect", "Unsupported Server : wrong ODS version");
	}

	// Now, verify the GDS32.DLL we are using is compatible with the server
	if (ODS == 10 && GDS::gds()->_GDSVersion < 60)
	{
		status.Reset();
	    (*GDS::gds()->_detach_database)(status.Self(), &_Handle);
        _Handle = NULL;     // Should be, but better be sure...
		throw ExceptionImpl("Database::Connect", "GDS32.DLL version 5 against IBSERVER 6");
	}

	/*
	if (ODS == 9 && GDS::gds()->_GDSVersion >= 60)
	{
		status.Reset();
	    (*GDS::gds()->_detach_database)(status.Self(), &_Handle);
        _Handle = NULL;     // Should be, but better be sure...
		throw ExceptionImpl("Database::Connect", "GDS32.DLL version 6 against IBSERVER 5");
	}
	*/
}

void DatabaseImpl::Disconnect(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Database::Disconnect", "Database is not connected.");

    _IBS status;

	// Cancel all pending event traps
	ClearEvents();

    // Rollback any started transaction...
	for (int i = 0; i < _TrCount; i++)
		try { _Transactions[i]->Rollback(); } catch (IBPP::Exception&) { }

	// Detach from the database
    (*GDS::gds()->_detach_database)(status.Self(), &_Handle);
    if (status.Errors())
		throw ExceptionImpl(&status, "Database::Disconnect", "isc_detach_database failed");

    _Handle = NULL;
}

void DatabaseImpl::Drop(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Database::Drop", "Database is not connected.");

	_IBS vector;

    // Annule toutes les transactions dans lesquelles nous participions...
	for (int i = 0; i < _TrCount; i++)
		try { _Transactions[i]->Rollback(); }
		catch (IBPP::Exception&) { }

	(*GDS::gds()->_drop_database)(vector.Self(), &_Handle);
    if (vector.Errors())
    	throw ExceptionImpl(&vector, "Database::Drop", "isc_drop_database failed");

    _Handle = NULL;
}

void DatabaseImpl::DefineEvent(const char* eventname, IBPP::EventInterface* objref)
{
	if (eventname == NULL || objref == NULL)
		throw ExceptionImpl("Database::DefineEvent", "NULL pointer reference detected.");

	if (_Events == NULL) _Events = new _EPB;
	else CancelEvents();

	_Events->Define(eventname, objref);
}

void DatabaseImpl::DropEvent(const char* eventname)
{
	if (_Events == NULL) throw ExceptionImpl("Database::DropEvent",
		"Attempt to drop an event while none are defined.");

	CancelEvents();
	_Events->Drop(eventname);
}

void DatabaseImpl::ClearEvents(void)
{
	CancelEvents();

	if (_Events != NULL)
	{
		delete _Events;
		_Events = NULL;
		_EventsId = 0;
	}
}

void DatabaseImpl::DispatchEvents(void)
{
	// If no events registered, nothing to do of course.
	// If we are still waiting for some events to fire, nothing to do, too.
	if (_Events == NULL || _EventsActive) return;

	if (_Handle == NULL)
		throw ExceptionImpl("Database::DispatchEvents", "Database is not connected.");

	// Let's fire the events actions for all the events which triggered.
	if (_EventsTrapped) _Events->FireActions(this);

	// Let's kick or re-kick the asynchronous wait
	_IBS vector;

	_EventsTrapped = false;
	_EventsActive = true;
	(*GDS::gds()->_que_events)(vector.Self(), &_Handle, &_EventsId,
		_Events->Size(), _Events->EventsBuffer(), (isc_callback)EventHandler, (char*)this);

    if (vector.Errors())
	{
		_EventsId = 0;	// Should be, but better be safe
		_EventsActive = false;
    	throw ExceptionImpl(&vector, "Database::DispatchEvents",
    		"isc_que_events failed");
	}
}

void DatabaseImpl::GetInfo(int* ODSMajor, int* ODSMinor,
	int* PageSize, bool* SyncWrites)
{
	char items[] = {isc_info_ods_version,
					isc_info_ods_minor_version,
					isc_info_page_size,
					isc_info_forced_writes,
					isc_info_end};
    _IBS status;
	_RB result(100);

	status.Reset();
	(*GDS::gds()->_database_info)(status.Self(), &_Handle, sizeof(items), items,
		result.Size(), result.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Database::GetInfo", "isc_database_info failed");

	if (ODSMajor != NULL) *ODSMajor = result.GetValue(isc_info_ods_version);
	if (ODSMinor != NULL) *ODSMinor = result.GetValue(isc_info_ods_minor_version);
	if (PageSize != NULL) *PageSize = result.GetValue(isc_info_page_size);
	if (SyncWrites != NULL)
		*SyncWrites = result.GetValue(isc_info_forced_writes) == 1 ? true : false;
}

void DatabaseImpl::GetStat(int* Pages, int* Memory, int* Buffers,
	int* Fetches, int* Marks, int* Reads, int* Writes)
{
	char items[] = {isc_info_allocation,
					isc_info_current_memory,
					isc_info_num_buffers,
					isc_info_fetches,
					isc_info_marks,
					isc_info_reads,
					isc_info_writes,
					isc_info_end};
    _IBS status;
	_RB result(100);

	status.Reset();
	(*GDS::gds()->_database_info)(status.Self(), &_Handle, sizeof(items), items,
		result.Size(), result.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Database::GetStat", "isc_database_info failed");

	if (Pages != NULL) *Pages = result.GetValue(isc_info_allocation);
	if (Memory != NULL) *Memory = result.GetValue(isc_info_current_memory);
	if (Buffers != NULL) *Buffers = result.GetValue(isc_info_num_buffers);
	if (Fetches != NULL) *Fetches = result.GetValue(isc_info_fetches);
	if (Marks != NULL) *Marks = result.GetValue(isc_info_marks);
	if (Reads != NULL) *Reads = result.GetValue(isc_info_reads);
	if (Writes != NULL) *Writes = result.GetValue(isc_info_writes);
}

void DatabaseImpl::Release(IBPP::Database*& Self)
{
	if (this != dynamic_cast<DatabaseImpl*>(Self))
		throw ExceptionImpl("Database::Release", "Invalid Release()");
	delete this;
	Self = NULL;
}

//	(((((((( OBJECT INTERNAL METHODS ))))))))

void DatabaseImpl::Init(void)
{
    _Handle = NULL;
	_Dialect = 3;

	_ServerName = NULL;
    _DatabaseName = NULL;

    _UserName = new char[7];	// "SYSDBA" + '\0'
	strcpy(_UserName, "SYSDBA");

    _UserPassword = new char[10];	// "masterkey" + '\0'
	strcpy(_UserPassword, "masterkey");

    _RoleName = NULL;
	_CreateParams = NULL;
	_TrCount = 0;
	_Transactions = NULL;
	_StCount = 0;
	_Statements = NULL;
	_BbCount = 0;
	_Blobs = NULL;
	_ArCount = 0;
	_Arrays = NULL;

	_Events = NULL;
	_EventsId = 0;
	_EventsActive = false;
	_EventsTrapped = false;
}

void DatabaseImpl::CancelEvents(void)
{
	if (_Events != NULL && _EventsActive)
	{
		if (_Handle == NULL) throw ExceptionImpl("Database::CancelEvents",
			"Database is not connected");

		_IBS vector;

		_EventsActive = false;
		(*GDS::gds()->_cancel_events)(vector.Self(), &_Handle, &_EventsId);

		_EventsId = 0;	// Should be, but better be safe
	    if (vector.Errors())
		{
	    	throw ExceptionImpl(&vector, "Database::CancelEvents",
	    		"isc_cancel_events failed");
		}
	}
}

void DatabaseImpl::EventUpdateCounts(short size, const char* tmpbuffer)
{
	if (size != _Events->Size())
		throw ExceptionImpl("Database::EventUpdateCount",
			"Unexpected temporary events buffer size != events result buffer size");

	for (int i = 0; i < _Events->Size(); i++)
		_Events->ResultsBuffer()[i] = tmpbuffer[i];

	_EventsTrapped = true;
}

void DatabaseImpl::EventHandler(const char* object,
	short size, const char* tmpbuffer)
{
	// >>>>> This method is a STATIC member !! <<<<<

	DatabaseImpl* db = (DatabaseImpl*)object;

	if (size != 0 && tmpbuffer != NULL)
		db->EventUpdateCounts(size, tmpbuffer);

	db->_EventsActive = false;
}

DatabaseImpl::DatabaseImpl(const char* ServerName, const char* DatabaseName,
    		const char* UserName, const char* UserPassword, const char* RoleName)
{
	Init();
	SetServerName(ServerName);
	SetDatabaseName(DatabaseName);
	SetUsername(UserName);
	SetUserPassword(UserPassword);
	SetRoleName(RoleName);
}

DatabaseImpl::~DatabaseImpl()
{
	try { Disconnect(); } catch (IBPP::Exception&) { }

	if (_ServerName != NULL) delete [] _ServerName;
	if (_DatabaseName != NULL) delete [] _DatabaseName;
	if (_UserName != NULL) delete [] _UserName;
	if (_UserPassword != NULL) delete [] _UserPassword;
	if (_RoleName != NULL) delete [] _RoleName;
	if (_CreateParams != NULL) delete [] _CreateParams;

	// Dtachons cet DatabaseImpl de tout IBlob. C'est trs important afin
	// que les objets IBlobImpl n'aient plus de rfrence  un DatabaseImpl qui
	// n'existe plus.
	// Balayage inverse du tableau pour viter effets de bords et amliorer performance.
	// En effet le tableau rtrci  chaque tour de boucle (_BbCount diminue).
	// Et lors de la suppression, il y compactage du tableau par copie d'lments
	// du haut vers le bas.
	int i;
	for (i = _BbCount-1; i >= 0; i--)
		try { _Blobs[i]->DetachDatabase(); }
			catch (IBPP::Exception&) { }

	// Dtachons cet DatabaseImpl de tout Array. C'est trs important afin
	// que les objets ArrayImpl n'aient plus de rfrence  un DatabaseImpl qui
	// n'existe plus.
	for (i = _ArCount-1; i >= 0; i--)
		try { _Arrays[i]->DetachDatabase(); }
			catch (IBPP::Exception&) { }

	// Dtachons cet DatabaseImpl de tout IStatementImpl. C'est trs important afin
	// que les objets IStatementImpl n'aient plus de rfrence  un DatabaseImpl qui
	// n'existe plus.
	for (i = _StCount-1; i >= 0; i--)
		try { _Statements[i]->DetachDatabase(); }
			catch (IBPP::Exception&) { }

	// Dtachons cet DatabaseImpl de toute ITransactionImpl. C'est trs important afin
	// que les objets ITransation n'aient plus de rfrence  un DatabaseImpl qui
	// n'existe plus.
	for (i = _TrCount-1; i >= 0; i--)
		try { _Transactions[i]->DetachDatabase(this); }
			catch (IBPP::Exception&) { }

	// Finalement librons les tableaux
	if (_Arrays != NULL) delete [] _Arrays;
	if (_Blobs != NULL) delete [] _Blobs;
	if (_Statements != NULL) delete [] _Statements;
	if (_Transactions != NULL) delete [] _Transactions;
}

//
//	EOF
//
