///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: date.cpp,v 1.5 2001/03/04 00:23:54 themule Exp $
//	Subject : IBPP, Date class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

#ifdef IBPP_LINUX

void GetLocalTime (SYSTEMTIME* st)
{
	time_t _time;
	time (&_time);
	struct tm* _tm = localtime (&_time);

	st->wYear = _tm->tm_year + 1900;
	st->wMonth = _tm->tm_mon + 1;
	st->wDayOfWeek = _tm->tm_wday;
	st->wDay = _tm->tm_mday;
	st->wHour = _tm->tm_hour;
	st->wMinute = _tm->tm_min;
	st->wSecond = _tm->tm_sec;
	/* TODO provide milliseconds */
	st->wMilliseconds = 0;
}

#endif

void IBPP::Date::Today(void)
{
	SYSTEMTIME st;
	GetLocalTime(&st);

	IBPP::itod(&_Date, (int)st.wYear, (int)st.wMonth, (int)st.wDay);
}

void IBPP::Date::SetDate(int year, int month, int day)
{
	IBPP::itod(&_Date, year, month, day);
	if (_Date < 1) throw ExceptionHelper("Date::SetDate", "Invalid year, month, day values");
}

void IBPP::Date::GetDate(int& year, int& month, int& day)
{
	IBPP::dtoi(_Date, &year, &month, &day);
}

void IBPP::Date::Add(int days)
{
	int year;
	int newdate = _Date + days;		// days can be signed
	IBPP::dtoi(newdate, &year, NULL, NULL);
	if (year == 0) throw ExceptionHelper("Date::Add()", "Invalid resulting date");
	_Date = newdate;
}

void IBPP::Date::StartOfMonth(void)
{
	int year, month;
	IBPP::dtoi(_Date, &year, &month, NULL);
	IBPP::itod(&_Date, year, month, 1);		// First of same month
}

void IBPP::Date::EndOfMonth(void)
{
	int year, month;
	IBPP::dtoi(_Date, &year, &month, NULL);
	if (++month > 12) { month = 1; year++; }
	IBPP::itod(&_Date, year, month, 1);		// First of next month
	_Date--;	// Last day of original month, all weird cases accounted for
}

IBPP::Date::Date(const int& copied)
{
	int year;
	IBPP::dtoi(copied, &year, NULL, NULL);
	if (year == 0) throw ExceptionHelper("Date::Date(int&)", "Invalid date");
	_Date = copied;
}

IBPP::Date::Date(int year, int month, int day)
{
	SetDate(year, month, day);
}

IBPP::Date::Date(const IBPP::Date& copied)
{
	_Date = copied._Date;
}

IBPP::Date& IBPP::Date::operator=(const IBPP::Date& assigned)
{
	_Date = assigned._Date;
	return *this;
}

IBPP::Date& IBPP::Date::operator=(const int& assigned)
{
	int year;
	IBPP::dtoi(assigned, &year, NULL, NULL);
	if (year == 0) throw ExceptionHelper("Date::operator=(int)", "Invalid date");
	_Date = assigned;
	return *this;
}

//	Eof
