///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: dbkey.cpp,v 1.6 2001/05/22 22:26:08 epocman Exp $
//	Subject : IBPP, DBKey class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

//	Private implementation

void IBPP::DBKey::BlindCopy(const DBKey& copied)
{
	_Size = copied._Size;

	if (copied._DBKey != NULL)
	{
		_DBKey = new char[_Size];
		memcpy(_DBKey, copied._DBKey, _Size);
	}
	else _DBKey = NULL;

	if (copied._String != NULL)
	{
		_String = new char[strlen(copied._String)+1];
		strcpy(_String, copied._String);
	}
	else _String = NULL;
}

//	Public implementation

void IBPP::DBKey::SetKey(const char* key, int size)
{
	if (key == NULL)
		throw ExceptionImpl("IBPP::DBKey::SetKey", "NULL DBKey reference detected.");
	if (size <= 0 || ((size >> 3) << 3) != size)
		throw ExceptionImpl("IBPP::DBKey::SetKey", "Invalid DBKey size.");

	if (_String != NULL)
	{
		delete [] _String;
		_String = NULL;
	}
	if (_DBKey != NULL) delete [] _DBKey;
	_DBKey = new char[size];
	memcpy(_DBKey, key, size);
	_Size = size;
}

void IBPP::DBKey::GetKey(char* key, int size)
{
	if (_DBKey == NULL)
		throw ExceptionImpl("IBPP::DBKey::GetKey", "DBKey not assigned.");
	if (key == NULL)
		throw ExceptionImpl("IBPP::DBKey::GetKey", "NULL DBKey reference detected.");
	if (size != _Size)
		throw ExceptionImpl("IBPP::DBKey::GetKey", "Incompatible DBKey size detected.");

	memcpy(key, _DBKey, size);
}

const char* IBPP::DBKey::AsString(void)
{
	if (_DBKey == NULL)
		throw ExceptionImpl("IBPP::DBKey::GetString", "DBKey not assigned.");

	if (_String == NULL)
	{
		char tmp[14];
		int n = _Size / 8;
		unsigned int * key = (unsigned int *)_DBKey;
		_String = new char[n*13 + (n-1) + 1];	// "4:8-4:8....\0"
		_String[0] = '\0';
		for (int i = 0; i < n; i++)
		{
			if (i != 0) strcat(_String, "-");
			sprintf(tmp, "%4.4x:%8.8x", (unsigned short)key[i*2], key[i*2+1]);
			strcat(_String, tmp);
		}
	}

	return _String;
}

IBPP::DBKey::DBKey()
{
	_DBKey = NULL;
	_String = NULL;
	_Size = 0;
}

IBPP::DBKey::DBKey(const DBKey& copied)
{
	BlindCopy(copied);
}

IBPP::DBKey& IBPP::DBKey::operator=(const IBPP::DBKey& assigned)
{
	if (_DBKey != NULL) delete [] _DBKey;
	if (_String != NULL) delete [] _String;
	BlindCopy(assigned);
	return *this;
}

IBPP::DBKey::~DBKey()
{
	if (_DBKey != NULL) delete [] _DBKey;
	if (_String != NULL) delete [] _String;
}

//
//	EOF
//
