///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: exceptionhelper.cpp,v 1.3 2001/05/22 22:26:08 epocman Exp $
//	Subject : IBPP, Initialization of the library
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* ExceptionHelper is a C++ exception class. It is always thrown by reference
//	  from inside the implementation of IBPP. It is always catched by reference
//	  as a IBPP::Exception class in code that use the IBPP library. The 'user'
//	  code does not have the ability to throw that exception.
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

//	(((((((( OBJECT INTERFACE IMPLEMENTATION ))))))))

const char* ExceptionHelper::GetOrigin(void)
{
	return _Origin;
}

void ExceptionHelper::GetErrorMessage(char* message)
{
	strcpy(message, "*** IBPP::Exception");
	if (_Origin != NULL)
	{
		strcat(message, " inside ");
		strcat(message, _Origin);
	}
	strcat(message, " ***\n\n");
	if (_Message != NULL)
	{
		strcat(message, "IBPP : ");
		strcat(message, _Message);
		strcat(message, "\n\n");
	}
}

void ExceptionHelper::PrintfErrorMessage(void)
{
	printf("*** IBPP::Exception");
	if (_Origin != NULL)  printf(" inside %s", _Origin);
	printf(" ***\n\n");
	if (_Message != NULL) printf("IBPP : %s\n\n", _Message);
}

//	(((((((( OBJECT INTERNAL METHODS ))))))))

void ExceptionHelper::BlindCopy(const ExceptionHelper& copied)
{
	if (copied._Origin != NULL)
	{
		_Origin = new char[strlen(copied._Origin)+1];
		strcpy(_Origin, copied._Origin);
	}
	else _Origin = NULL;

	if (copied._Message != NULL)
	{
		_Message = new char[strlen(copied._Message)+1];
		strcpy(_Message, copied._Message);
	}
	else _Message = NULL;
}

void ExceptionHelper::SetOrigin(const char* origin)
{
	if (_Origin != NULL) delete [] _Origin;
	if (origin == NULL) _Origin = NULL;
	else
	{
		_Origin = new char[strlen(origin)+1];
		strcpy(_Origin, origin);
	}
}

void ExceptionHelper::SetMessage(const char* message)
{
	if (_Message != NULL) delete [] _Message;
	if (message == NULL) _Message = NULL;
	else
	{
		_Message = new char[strlen(message)+1];
		strcpy(_Message, message);
	}
}

ExceptionHelper::ExceptionHelper()
{
	_Origin = NULL;
	_Message = NULL;
}

ExceptionHelper::ExceptionHelper(const ExceptionHelper& copied)
{
	BlindCopy(copied);
}

ExceptionHelper& ExceptionHelper::operator=(const ExceptionHelper& copied)
{
	//	The copy constructor is VERY important for a class used
	//	as a C++ exception, especially with Borland BCC32 5.5

	if (_Origin != NULL) delete [] _Origin;
	if (_Message != NULL) delete [] _Message;
	BlindCopy(copied);
	return *this;
}

ExceptionHelper::ExceptionHelper(const char* origin, const char* message)
{
	_Origin = NULL;
	_Message = NULL;
	SetOrigin(origin);
	SetMessage(message);
}

ExceptionHelper::~ExceptionHelper()
{
	if (_Origin != NULL) delete [] _Origin;
	if (_Message != NULL) delete [] _Message;
}

//
//	EOF
//

