///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: ibpp.cpp,v 1.10 2001/11/03 11:13:17 epocman Exp $
//	Subject : IBPP, Initialization of the library
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

using namespace ibpp_internals;

GDS* GDS::Instance = NULL;

GDS* GDS::gds(void)
{
	// Let's load the GDS32.DLL library, if it is not already loaded.
	// The load is guaranteed to be done only once per application.

	if (Instance == NULL)
	{
		// Load the GDS32.DLL library
		Instance = new GDS();
		if (Instance == NULL)
			throw ExceptionImpl("GDS::gds()", "Memory allocation problem");

#ifdef IBPP_WINDOWS
		Instance->_Handle = LoadLibrary("GDS32.DLL");
		if (Instance->_Handle == NULL)
		{
			delete Instance;
			Instance = NULL;
			throw ExceptionImpl("GDS::gds()", "Could not find or load the GDS32.DLL");
		}
#endif

		Instance->_GDSVersion = 60;

		// Get the entry points that we need

#ifdef IBPP_WINDOWS
#define IB_ENTRYPOINT(X) \
			if ((Instance->_##X = (proto_##X*)GetProcAddress(Instance->_Handle, "isc_"#X)) == NULL) \
				throw ExceptionImpl("GDS:gds()", "Entry-point isc_"#X" not found")
#elif defined IBPP_LINUX
/* TODO : perform a late-bind on linux */					
#define IB_ENTRYPOINT(X) Instance->_##X = (proto_##X*)isc_##X 
#endif					

		IB_ENTRYPOINT(attach_database);
		IB_ENTRYPOINT(detach_database);
		IB_ENTRYPOINT(drop_database);
		IB_ENTRYPOINT(database_info);
		IB_ENTRYPOINT(open_blob2);
		IB_ENTRYPOINT(create_blob);
		IB_ENTRYPOINT(close_blob);
		IB_ENTRYPOINT(cancel_blob);
		IB_ENTRYPOINT(get_segment);
		IB_ENTRYPOINT(put_segment);
		IB_ENTRYPOINT(blob_info);
		IB_ENTRYPOINT(vax_integer);
		IB_ENTRYPOINT(sqlcode);
		IB_ENTRYPOINT(sql_interprete);
		IB_ENTRYPOINT(interprete);
		IB_ENTRYPOINT(que_events);
		IB_ENTRYPOINT(cancel_events);
		IB_ENTRYPOINT(start_multiple);
		IB_ENTRYPOINT(commit_transaction);
		IB_ENTRYPOINT(commit_retaining);
		IB_ENTRYPOINT(rollback_transaction);
		IB_ENTRYPOINT(rollback_retaining);
		IB_ENTRYPOINT(dsql_execute_immediate);
		IB_ENTRYPOINT(dsql_allocate_statement);
		IB_ENTRYPOINT(dsql_describe);
		IB_ENTRYPOINT(dsql_describe_bind);
		IB_ENTRYPOINT(dsql_prepare);
		IB_ENTRYPOINT(dsql_execute);
		IB_ENTRYPOINT(dsql_execute2);
		IB_ENTRYPOINT(dsql_fetch);
		IB_ENTRYPOINT(dsql_free_statement);
		IB_ENTRYPOINT(dsql_set_cursor_name);
		IB_ENTRYPOINT(dsql_sql_info);
		//IB_ENTRYPOINT(decode_date);
		//IB_ENTRYPOINT(encode_date);
		//IB_ENTRYPOINT(add_user);
		//IB_ENTRYPOINT(modify_user);
		//IB_ENTRYPOINT(delete_user);

		IB_ENTRYPOINT(service_attach);
		IB_ENTRYPOINT(service_detach);
		IB_ENTRYPOINT(service_start);
		IB_ENTRYPOINT(service_query);
		IB_ENTRYPOINT(decode_sql_date);
		IB_ENTRYPOINT(decode_sql_time);
		IB_ENTRYPOINT(decode_timestamp);
		IB_ENTRYPOINT(encode_sql_date);
		IB_ENTRYPOINT(encode_sql_time);
		IB_ENTRYPOINT(encode_timestamp);
	}

	return Instance;
}

void GDS::Release(void)
{
	if (Instance != NULL)
	{
#ifdef IBPP_WINDOWS		
		if (Instance->_Handle != NULL)
		{
			FreeLibrary(Instance->_Handle);
			Instance->_Handle = NULL;
		}
#endif		
		delete Instance;
		Instance = NULL;
	}
}

bool IBPP::CheckVersion(unsigned long AppVersion)
{
	(void) GDS::gds();			// Just call it to trigger the initialization
	return (AppVersion & 0xFFFFFF00) ==
			(IBPP::Version & 0xFFFFFF00) ? true : false;
}

int IBPP::GDSVersion(void)
{
	return GDS::gds()->_GDSVersion;
}

void IBPP::Release(void)
{
	GDS::Release();
}

//	Factories for our Interface objects

IBPP::Service* IBPP::ServiceFactory(const char* ServerName,
			const char* UserName, const char* UserPassword)
{
	(void) GDS::gds();			// Triggers the initialization, if needed
	ServiceImpl* impl = new ServiceImpl(ServerName, UserName, UserPassword);
	if (impl == NULL) throw ExceptionImpl("IBPP::ServiceFactory", "Memory allocation error");
	return impl;
}

IBPP::Database* IBPP::DatabaseFactory(const char* ServerName, const char* DatabaseName,
			const char* UserName, const char* UserPassword, const char* RoleName)
{
	(void) GDS::gds();			// Triggers the initialization, if needed
	DatabaseImpl* impl = new DatabaseImpl(ServerName, DatabaseName, UserName,
		UserPassword, RoleName);
	if (impl == NULL) throw ExceptionImpl("IBPP::DatabaseFactory", "Memory allocation error");
	return impl;
}

IBPP::Transaction* IBPP::TransactionFactory(Database* db, TAM am, TIL il, TLR lr)
{
	(void) GDS::gds();			// Triggers the initialization, if needed
	DatabaseImpl* dbimpl = dynamic_cast<DatabaseImpl*>(db);
	TransactionImpl* impl = new TransactionImpl(dbimpl, am, il, lr);
	if (impl == NULL) throw ExceptionImpl("IBPP::TransactionFactory", "Memory allocation error");
	return impl;
}

IBPP::Statement* IBPP::StatementFactory(Database* db, Transaction* tr, const char* sql)
{
	(void) GDS::gds();			// Triggers the initialization, if needed
	DatabaseImpl* dbimpl = dynamic_cast<DatabaseImpl*>(db);
	TransactionImpl* trimpl = dynamic_cast<TransactionImpl*>(tr);
	StatementImpl* impl = new StatementImpl(dbimpl, trimpl, sql);
	if (impl == NULL) throw ExceptionImpl("IBPP::StatementFactory", "Memory allocation error");
	return impl;
}

IBPP::Blob* IBPP::BlobFactory(Database* db, Transaction* tr)
{
	(void) GDS::gds();	// Triggers the library initialization, if needed
	DatabaseImpl* dbimpl = dynamic_cast<DatabaseImpl*>(db);
	TransactionImpl* trimpl = dynamic_cast<TransactionImpl*>(tr);
	BlobImpl* impl = new BlobImpl(dbimpl, trimpl);
	if (impl == NULL) throw ExceptionImpl("IBPP::BlobFactory", "Memory allocation error");
	return impl;
}

IBPP::Array* IBPP::ArrayFactory(Database* db, Transaction* tr)
{
	(void) GDS::gds();	// Triggers the library initialization, if needed
	DatabaseImpl* dbimpl = dynamic_cast<DatabaseImpl*>(db);
	TransactionImpl* trimpl = dynamic_cast<TransactionImpl*>(tr);
	ArrayImpl* impl = new ArrayImpl(dbimpl, trimpl);
	if (impl == NULL) throw ExceptionImpl("IBPP::ArrayFactory", "Memory allocation error");
	return impl;
}

//
//	EOF
//
