///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: service.cpp,v 1.7 2001/11/03 21:01:43 epocman Exp $
//	Subject : IBPP, Service class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

#ifdef IBPP_LINUX
#include <unistd.h>
#endif

using namespace ibpp_internals;

//	(((((((( OBJECT INTERFACE IMPLEMENTATION ))))))))

void ServiceImpl::Connect(void)
{
	if (GDS::gds()->_GDSVersion < 60)
		throw ExceptionImpl("Service", "Requires the version 6 of GDS32.DLL");
	if (_Handle	!= NULL)
		throw ExceptionImpl("Service::Connect", "Service	is already connected.");
	if (_UserName == NULL)
		throw ExceptionImpl("Service::Connect", "Unspecified user name.");
	if (_UserPassword == NULL)
		throw ExceptionImpl("Service::Connect", "Unspecified user password.");

	// Attach to the Service Manager
	_IBS status;
	_SPB spb;
	char connect[128];

	// Build a SPB based on	the	properties
	spb.Insert(isc_spb_version);
	spb.Insert(isc_spb_current_version);
	spb.Insert(isc_spb_user_name, 1, _UserName);
	spb.Insert(isc_spb_password, 1,	_UserPassword);

	if (_ServerName	!= NULL)
	{
		strcpy(connect,	_ServerName);
		strcat(connect,	":");
	}
	else connect[0]	= '\0';
	strcat(connect,	"service_mgr");

	(*GDS::gds()->_service_attach)(status.Self(), (short)strlen(connect), connect,
		&_Handle, spb.Size(), spb.Self());
	if (status.Errors())
	{
		_Handle	= NULL;		// Should be, but better be	sure...
		throw ExceptionImpl(&status, "Service::Connect", "isc_service_attache	failed");
	}
}

void ServiceImpl::Disconnect(void)
{
	if (GDS::gds()->_GDSVersion < 60)
		throw ExceptionImpl("Service", "Requires the version 6 of GDS32.DLL");
	if (_Handle	== NULL)
		throw ExceptionImpl("Service::Disconnect", "Service is not connected.");

	_IBS status;

	// Detach from the service manager
	(*GDS::gds()->_service_detach)(status.Self(), &_Handle);
	if (status.Errors())
		throw ExceptionImpl(&status, "Service::Disconnect", "isc_service_detach failed");

	_Handle	= NULL;
}

void ServiceImpl::AddUser(const char* username, const char* password,
			const char* first, const char* middle, const char* last)
{
	if (GDS::gds()->_GDSVersion >= 60 && _Handle == NULL)
		throw ExceptionImpl("Service::AddUser", "Service	is not connected.");
	if (username ==	NULL ||	strlen(username) ==	0)
		throw ExceptionImpl("Service::AddUser", "Username	required.");
	if (password ==	NULL ||	strlen(password) ==	0)
		throw ExceptionImpl("Service::AddUser", "Password	required.");

	_IBS status;
	_SPB spb;
	spb.Insert(isc_action_svc_add_user);
	spb.Insert(isc_spb_sec_username, 2,	username);
	spb.Insert(isc_spb_sec_password, 2,	password);
	if (first != NULL) spb.Insert(isc_spb_sec_firstname, 2,	first);
	if (middle != NULL)	spb.Insert(isc_spb_sec_middlename, 2, middle);
	if (last !=	NULL) spb.Insert(isc_spb_sec_lastname, 2, last);

	(*GDS::gds()->_service_start)(status.Self(), &_Handle, NULL, spb.Size(), spb.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Service::AddUser", "isc_service_start failed");

	Wait(NULL);
}

void ServiceImpl::ModifyUser(const char* username, const char* password,
			const char* first, const char* middle, const char* last)
{
	if (GDS::gds()->_GDSVersion >= 60 && _Handle == NULL)
		throw ExceptionImpl("Service::ModifyUser", "Service	is not connected.");
	if (username ==	NULL ||	strlen(username) ==	0)
		throw ExceptionImpl("Service::ModifyUser", "Username required.");

	_IBS status;
	_SPB spb;

	spb.Insert(isc_action_svc_modify_user);
	spb.Insert(isc_spb_sec_username, 2,	username);
	if (password !=	NULL) spb.Insert(isc_spb_sec_password, 2, password);
	if (first != NULL) spb.Insert(isc_spb_sec_firstname, 2,	first);
	if (middle != NULL)	spb.Insert(isc_spb_sec_middlename, 2, middle);
	if (last !=	NULL) spb.Insert(isc_spb_sec_lastname, 2, last);

	(*GDS::gds()->_service_start)(status.Self(), &_Handle, NULL, spb.Size(), spb.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Service::ModifyUser", "isc_service_start failed");

	Wait(NULL);
}

void ServiceImpl::RemoveUser(const char* username)
{

	if (GDS::gds()->_GDSVersion >= 60 && _Handle == NULL)
		throw ExceptionImpl("Service::AddUser", "Service	is not connected.");
	if (username ==	NULL ||	strlen(username) ==	0)
		throw ExceptionImpl("Service::AddUser", "Service	is not connected.");

	_IBS status;
	_SPB spb;

	spb.Insert(isc_action_svc_delete_user);
	spb.Insert(isc_spb_sec_username, 2,	username);

	(*GDS::gds()->_service_start)(status.Self(), &_Handle, NULL, spb.Size(), spb.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Service::RemoveUser", "isc_service_start failed");

	Wait(NULL);
}

void ServiceImpl::StartBackup(const char* dbfile, const char* bkfile, bool verbose)
{
	if (GDS::gds()->_GDSVersion < 60)
		throw ExceptionImpl("Service", "Requires the version 6 of GDS32.DLL");
	if (_Handle	== NULL)
		throw ExceptionImpl("Service::Backup", "Service is not connected.");
	if (dbfile == NULL || strlen(dbfile) ==	0)
		throw ExceptionImpl("Service::Backup", "Main database	file must be specified.");
	if (bkfile == NULL || strlen(bkfile) ==	0)
		throw ExceptionImpl("Service::Backup", "Backup file must be specified.");

	_IBS status;
	_SPB spb;

	spb.Insert(isc_action_svc_backup);
	if (verbose) spb.Insert(isc_spb_verbose);
	spb.Insert(isc_spb_dbname, 2, dbfile);
	spb.Insert(isc_spb_bkp_file, 2,	bkfile);

	(*GDS::gds()->_service_start)(status.Self(), &_Handle, NULL, spb.Size(), spb.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Service::Backup", "isc_service_start failed");
}

void ServiceImpl::StartRestore(const char* bkfile, const char* dbfile,
			int	pagesize, bool verbose)
{
	if (GDS::gds()->_GDSVersion < 60)
		throw ExceptionImpl("Service", "Requires the version 6 of GDS32.DLL");
	if (_Handle	== NULL)
		throw ExceptionImpl("Service::Restore", "Service	is not connected.");
	if (bkfile == NULL || strlen(bkfile) ==	0)
		throw ExceptionImpl("Service::Restore", "Backup file must	be specified.");
	if (dbfile == NULL || strlen(dbfile) ==	0)
		throw ExceptionImpl("Service::Restore", "Main	database file must be specified.");

	_IBS status;
	_SPB spb;

	spb.Insert(isc_action_svc_restore);
	if (verbose) spb.Insert(isc_spb_verbose);
	spb.Insert(isc_spb_options,	isc_spb_res_create);
	spb.Insert(isc_spb_bkp_file, 2,	bkfile);
	spb.Insert(isc_spb_dbname, 2, dbfile);
	if (pagesize !=	0) spb.Insert(isc_spb_res_page_size, pagesize);

	(*GDS::gds()->_service_start)(status.Self(), &_Handle, NULL, spb.Size(), spb.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Service::Restore", "isc_service_start failed");
}

bool ServiceImpl::Wait(char* line)
{
	_IBS status;
	_SPB spb;
	_RB rb;
	char msg[512];

	// Sleeps 1	millisecond. This will release the remaining timeslot of the
	// thread. Doing so	will give a	good chance	for	small services tasks to
	// finish before we	check if they are still	running.

	if (GDS::gds()->_GDSVersion < 60)
		throw ExceptionImpl("Service", "Requires the version 6 of GDS32.DLL");

#ifdef IBPP_WINDOWS	
	Sleep(1);
#elif IBPP_LINUX
	usleep(1);
#endif	

	spb.Insert(isc_info_svc_line);
	while (true)
	{
		(*GDS::gds()->_service_query)(status.Self(), &_Handle, NULL, 0,	NULL,
			spb.Size(),	spb.Self(),	rb.Size(), rb.Self());
		if (status.Errors())
			throw ExceptionImpl(&status, "ServiceImpl::Wait", "isc_service_query failed");

		// If message length is	zero bytes,	task is	finished
		if (rb.GetString(isc_info_svc_line,	msg) ==	0) return false;

		// If there	is report data,	task is	still in progress.
		// If we are interested	in report (line	!= NULL), let's	break the loop.
		if (line !=	NULL) break;

		// Not interested in the report, let's loop	until task finished.
		status.Reset();
		rb.Reset();
	}

	// Task	is not finished, but we	have something to report
	strcpy(line, msg);
	return true;
}

void ServiceImpl::Release(IBPP::Service*& Self)
{
	if (this != dynamic_cast<ServiceImpl*>(Self))
		throw ExceptionImpl("Service::Release", "Invalid Release()");
	delete this;
	Self = NULL;
}

//	(((((((( OBJECT INTERNAL METHODS ))))))))

void ServiceImpl::SetServerName(const char* newName)
{
	if (_ServerName	!= NULL) delete	[] _ServerName;
	if (newName	== NULL) _ServerName = NULL;
	else
	{
		_ServerName	= new char[strlen(newName)+1];
		strcpy(_ServerName,	newName);
	}
}

void ServiceImpl::SetUserName(const char* newName)
{
	if (_UserName != NULL) delete [] _UserName;
	if (newName	== NULL) _UserName = NULL;
	else
	{
		_UserName =	new	char[strlen(newName)+1];
		strcpy(_UserName, newName);
	}
}

void ServiceImpl::SetUserPassword(const char* newPassword)
{
	if (_UserPassword != NULL) delete [] _UserPassword;
	if (newPassword	== NULL) _UserPassword = NULL;
	else
	{
		_UserPassword =	new	char[strlen(newPassword)+1];
		strcpy(_UserPassword, newPassword);
	}
}

ServiceImpl::ServiceImpl(const char* ServerName,
				const char* UserName, const char* UserPassword)
{
	_Handle	= NULL;
	_ServerName	= NULL;
	_UserName =	NULL;
	_UserPassword =	NULL;

	SetServerName(ServerName);
	SetUserName(UserName);
	SetUserPassword(UserPassword);
}

ServiceImpl::~ServiceImpl()
{
	try	{ Disconnect();	}
		catch (IBPP::Exception&)	{ }

	if (_ServerName	!= NULL) delete	[] _ServerName;
	if (_UserName != NULL) delete [] _UserName;
	if (_UserPassword != NULL) delete [] _UserPassword;
}

//
//	Eof
//
