///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: statement.cpp,v 1.13 2001/11/03 21:01:43 epocman Exp $
//	Subject : IBPP, Service class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

//	(((((((( OBJECT INTERFACE IMPLEMENTATION ))))))))

void StatementImpl::AttachDatabase(IBPP::Database* database)
{
	if (database == NULL)
		throw ExceptionImpl("Statement::AttachDatabase",
			"Can't attach a NULL IDatabase object.");

	if (_Database != NULL) _Database->DetachStatement(this);
	_Database = dynamic_cast<DatabaseImpl*>(database);
	_Database->AttachStatement(this);
}

void StatementImpl::DetachDatabase(void)
{
	if (_Database == NULL)
		throw ExceptionImpl("Statement::DetachDatabase",
			"No IDatabase was attached.");

	try {Close();} catch(IBPP::Exception&) {}
	_Database->DetachStatement(this);
	_Database = NULL;
}

void StatementImpl::AttachTransaction(IBPP::Transaction* transaction)
{
	if (transaction == NULL)
		throw ExceptionImpl("Statement::AttachTransaction",
			"Can't attach a NULL ITransaction object.");

	if (_Transaction != NULL) _Transaction->DetachStatement(this);
	_Transaction = dynamic_cast<TransactionImpl*>(transaction);
	_Transaction->AttachStatement(this);
}

void StatementImpl::DetachTransaction(void)
{
	if (_Transaction == NULL)
		throw ExceptionImpl("Statement::DetachTransaction",
			"No ITransaction was attached.");

	try {Close();} catch(IBPP::Exception&) {}
	_Transaction->DetachStatement(this);
	_Transaction = NULL;
}

void StatementImpl::Prepare(const char* sql)
{
	if (_Database == NULL)
		throw ExceptionImpl("Statement::Prepare", "An IDatabase must be attached.");
	if (_Database->GetHandle() == NULL)
		throw ExceptionImpl("Statement::Prepare", "IDatabase must be connected.");
	if (_Transaction == NULL)
		throw ExceptionImpl("Statement::Prepare", "An ITransaction must be attached.");
	if (_Transaction->GetHandle() == NULL)
		throw ExceptionImpl("Statement::Prepare", "ITransaction must be started.");
	if (sql == NULL)
		throw ExceptionImpl("Statement::Prepare", "SQL statement can't be NULL.");

	_IBS status;

	// Libre toute ressource ventuellement associe  cet Statement, puis
	// alloue un nouveau descripteur.
	try {Close();} catch(IBPP::Exception&) {}
	(*GDS::gds()->_dsql_allocate_statement)(status.Self(), _Database->GetHandlePtr(), &_Handle);
	if (status.Errors())
		throw ExceptionImpl(&status, "Statement::Prepare",
			"isc_dsql_allocate_statement failed.");

	// Estimation empirique du nombre de paramtres et du nombre de valeurs
	// en retour de la requte. Une surestimation est prfrable. Ce qu'il faut
	// viter c'est d'avoir  r-allouer les descripteurs. Les allouer un peu
	// trop grand est vraiment bnin.
	short inEstimate = 10;
	short outEstimate = 10;
	/*
	for (int i = 0; i < (int)strlen(sql); i++)
	{
		if (sql[i] == '?') inEstimate++;
		if (sql[i] == ',') outEstimate++;
	}
	outEstimate++;
	if (inEstimate >= 1) outEstimate -= (short)(inEstimate - 1);
	else inEstimate = 1;
	*/

	// Allocation du descripteur d'output et prparation de la requte
	_OutDescr = new _XSQLDA(outEstimate);
	status.Reset();
	(*GDS::gds()->_dsql_prepare)(status.Self(), _Transaction->GetHandlePtr(),
		&_Handle, (short)strlen(sql), const_cast<char*>(sql), 3 /*_Database->GetDialect()*/, _OutDescr->Self());
	if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl(&status, "Statement::Prepare",
			"isc_dsql_prepare failed.");
	}

	// Obtention du type de requte prpare
	status.Reset();
	char itemsReq[] = {isc_info_sql_stmt_type};
	char itemsRes[8];
	(*GDS::gds()->_dsql_sql_info)(status.Self(), &_Handle, 1, itemsReq,
		sizeof(itemsRes), itemsRes);
	if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl(&status, "Statement::Prepare",
			"isc_dsql_sql_info failed.");
	}
	if (itemsRes[0] == isc_info_sql_stmt_type)
	{
		switch (itemsRes[3])
		{
			case isc_info_sql_stmt_select :		_Type = IBPP::stSelect; break;
			case isc_info_sql_stmt_insert :		_Type = IBPP::stInsert; break;
			case isc_info_sql_stmt_update :		_Type = IBPP::stUpdate; break;
			case isc_info_sql_stmt_delete :		_Type = IBPP::stDelete; break;
			case isc_info_sql_stmt_ddl :		_Type = IBPP::stDDL; break;
			case isc_info_sql_stmt_exec_procedure : _Type = IBPP::stExecProcedure; break;
			case isc_info_sql_stmt_select_for_upd : _Type = IBPP::stSelectUpdate; break;
			case isc_info_sql_stmt_set_generator :	_Type = IBPP::stOther; break;
			default : _Type = IBPP::stUnsupported;
		}
	}
	if (_Type == IBPP::stUnknown || _Type == IBPP::stUnsupported)
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl("Statement::Prepare",
			"Unknown or unsupported statement type");
	}

	if (_OutDescr->DescrSize() == 0)
	{
		// Elimine le _OutDescr s'il n'est pas ncessaire (pas d'output)
		delete _OutDescr;
		_OutDescr = NULL;
	}
	else if (_OutDescr->DescrSize() > _OutDescr->AllocatedSize())
	{
		// Redimensionne le _OutDescr, la requte ne doit pas etre
		// prpare  nouveau, mais il faut redcrire les colonnes.
		_OutDescr->Resize(_OutDescr->DescrSize());
		status.Reset();
		(*GDS::gds()->_dsql_describe)(status.Self(), &_Handle, 1, _OutDescr->Self());
		if (status.Errors())
		{
			try {Close();} catch(IBPP::Exception&) {}
			throw ExceptionImpl(&status, "Statement::Prepare",
				"isc_dsql_describe failed.");
		}
	}

	// Prparation du _InDescr
	_InDescr = new _XSQLDA(inEstimate);
	status.Reset();
	(*GDS::gds()->_dsql_describe_bind)(status.Self(), &_Handle, 1, _InDescr->Self());
	if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl(&status, "Statement::Prepare",
			"isc_dsql_describe_bind failed.");
	}

	if (_InDescr->DescrSize() == 0)
	{
		// Elimine le _InDescr s'il n'est pas ncessaire (pas d'input)
		delete _InDescr;
		_InDescr = NULL;
	}
	else if (_InDescr->DescrSize() > _InDescr->AllocatedSize())
	{
		// Redimensionne le _InDescr
		_InDescr->Resize(_InDescr->DescrSize());
		status.Reset();
		(*GDS::gds()->_dsql_describe_bind)(status.Self(), &_Handle, 1, _InDescr->Self());
		if (status.Errors())
		{
			try {Close();} catch(IBPP::Exception&) {}
			throw ExceptionImpl(&status, "Statement::Prepare",
				"isc_dsql_describe_bind failed.");
		}
	}

	// Alloue les variables ncessaires dans le _InDescr
	if (_InDescr != NULL)
	{
		_InDescr->AllocVariables();
		_InMissing = new bool[_InDescr->DescrSize()];
		for (int i = 0; i < _InDescr->DescrSize(); i++) _InMissing[i] = true;
	}

	// Alloue les variables ncessaires dans le _OutDescr
	if (_OutDescr != NULL) _OutDescr->AllocVariables();
}

void StatementImpl::Execute(const char* sql)
{
	if (sql != NULL) Prepare(sql);

	if (_Handle == NULL)
		throw ExceptionImpl("Statement::Execute", "No statement has been prepared.");

	if (_InDescr != NULL)
		for (int i = 0; i < _InDescr->DescrSize(); i++)
			if (_InMissing[i]) throw ExceptionImpl("Statement::Execute",
								"All parameters must be specified.");

	CursorFree();		// Libre un ventuel cursor prcdent

	_IBS status;
	if (_Type == IBPP::stSelect)
	{
		// Retourne un Result Set (en principe)
		(*GDS::gds()->_dsql_execute)(status.Self(), _Transaction->GetHandlePtr(),
			&_Handle, 1, _InDescr == NULL ? NULL : _InDescr->Self());
		if (status.Errors())
		{
			//Close();	Commented because Execute error should not free the statement
			throw ExceptionImpl(&status, "Statement::Execute",
				"isc_dsql_execute failed.");
		}
		if (_OutDescr != NULL) _ResultSetAvailable = true;
	}
	else
	{
		// Retourne une seule ligne de rsultats (voire une seule valeur)
		(*GDS::gds()->_dsql_execute2)(status.Self(), _Transaction->GetHandlePtr(),
			&_Handle, 1, _InDescr == NULL ? NULL : _InDescr->Self(),
			_OutDescr == NULL ? NULL : _OutDescr->Self());
		if (status.Errors())
		{
			//Close();	Commented because Execute error should not free the statement
			throw ExceptionImpl(&status, "Statement::Execute",
				"isc_dsql_execute2 failed.");
		}
	}
}

void StatementImpl::CursorExecute(const char* cursor, const char* sql)
{
	if (cursor == NULL)
		throw ExceptionImpl("Statement::CursorExecute", "Cursor name can't be NULL.");

	if (sql != NULL) Prepare(sql);

	if (_Handle == NULL)
		throw ExceptionImpl("Statement::CursorExecute", "No statement has been prepared.");
	if (_Type != IBPP::stSelectUpdate)
		throw ExceptionImpl("Statement::CursorExecute", "Statement must be a SELECT FOR UPDATE.");
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::CursorExecute", "Statement would return no rows.");

	if (_InDescr != NULL)
		for (int i = 0; i < _InDescr->DescrSize(); i++)
			if (_InMissing[i]) throw ExceptionImpl("Statement::CursorExecute",
								"All parameters must be specified.");

	CursorFree();		// Libre un ventuel cursor prcdent

	_IBS status;
	(*GDS::gds()->_dsql_execute)(status.Self(), _Transaction->GetHandlePtr(),
		&_Handle, 1, _InDescr == NULL ? NULL : _InDescr->Self());
	if (status.Errors())
	{
		//Close();	Commented because Execute error should not free the statement
		throw ExceptionImpl(&status, "Statement::CursorExecute",
			"isc_dsql_execute failed.");
	}

	status.Reset();
	(*GDS::gds()->_dsql_set_cursor_name)(status.Self(), &_Handle, const_cast<char*>(cursor), 0);
	if (status.Errors())
	{
		//Close();	Commented because Execute error should not free the statement
		throw ExceptionImpl(&status, "Statement::CursorExecute",
			"isc_dsql_set_cursor_name failed");
	}

	_ResultSetAvailable = true;
}

void StatementImpl::ExecuteImmediate(const char* sql)
{
	if (_Database == NULL)
		throw ExceptionImpl("Statement::ExecuteImmediate", "An IDatabase must be attached.");
	if (_Database->GetHandle() == NULL)
		throw ExceptionImpl("Statement::ExecuteImmediate", "IDatabase must be connected.");
	if (_Transaction == NULL)
		throw ExceptionImpl("Statement::ExecuteImmediate", "An ITransaction must be attached.");
	if (_Transaction->GetHandle() == NULL)
		throw ExceptionImpl("Statement::ExecuteImmediate", "ITransaction must be started.");
	if (sql == NULL)
		throw ExceptionImpl("Statement::ExecuteImmediate", "SQL statement can't be NULL.");

	_IBS status;
	try {Close();} catch(IBPP::Exception&) {}
    (*GDS::gds()->_dsql_execute_immediate)(status.Self(), _Database->GetHandlePtr(),
    	_Transaction->GetHandlePtr(), 0, const_cast<char*>(sql), 3 /*_Database->GetDialect()*/, NULL);
    if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl(&status, "Statement::ExecuteImmediate",
				"isc_dsql_execute_immediate failed.");
	}
}

int StatementImpl::AffectedRows(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::AffectedRows", "No statement has been prepared.");
	if (_Database == NULL)
		throw ExceptionImpl("Statement::AffectedRows", "A Database must be attached.");
	if (_Database->GetHandle() == NULL)
		throw ExceptionImpl("Statement::AffectedRows", "Database must be connected.");

	if (_Type != IBPP::stInsert && _Type != IBPP::stUpdate && _Type != IBPP::stDelete)
	{
		// We return a count of 0 for any non interesting request. For selects, it is
		// assumed to be better to count the rows as they are fetched. So this IBPP API
		// can only be called after a successfull INSERT, UPDATE or DELETE.
		return 0;
	}

	int count;
	_IBS status;
	_RB result;
	char itemsReq[] = {isc_info_sql_records};

	(*GDS::gds()->_dsql_sql_info)(status.Self(), &_Handle, 1, itemsReq,
		result.Size(), result.Self());
	if (status.Errors()) throw ExceptionImpl(&status,
			"Statement::AffectedRows", "isc_dsql_sql_info failed.");

	if (_Type == IBPP::stInsert)
			count = result.GetValue(isc_info_sql_records, isc_info_req_insert_count);
	else if (_Type == IBPP::stUpdate)
			count = result.GetValue(isc_info_sql_records, isc_info_req_update_count);
	else if (_Type == IBPP::stDelete)
			count = result.GetValue(isc_info_sql_records, isc_info_req_delete_count);
	else	count = 0;	// Should never happen

	return count;
}

bool StatementImpl::Fetch(void)
{
	if (! _ResultSetAvailable)
		throw ExceptionImpl("Statement::Fetch",
			"No statement has been executed or no result set available.");

	_IBS status;
	int code = (*GDS::gds()->_dsql_fetch)(status.Self(), &_Handle, 1, _OutDescr->Self());
	if (code == 100)	// This special code means "no more rows"
	{
		CursorFree();	// Free the explicit or implicit cursor/result-set
		return false;
	}
	if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl(&status, "Statement::Fetch",
			"isc_dsql_fetch failed.");
	}

	return true;
}

void StatementImpl::Close(void)
{
	// Libere toutes ressources. Utilis avant de prparer un nouveau
	// statement sur la meme IDatabase attache ou depuis le destructeur.
	if (_Handle != NULL)
	{
		_IBS status;
		(*GDS::gds()->_dsql_free_statement)(status.Self(), &_Handle, DSQL_drop);
		_Handle = NULL;
		if (status.Errors())
			throw ExceptionImpl(&status, "Statement::Close",
				"isc_dsql_free_statement failed.");
	}
	if (_InMissing != NULL)
	{
		delete [] _InMissing;
		_InMissing = NULL;
	}
	if (_InDescr != NULL)
	{
		delete _InDescr;
		_InDescr = NULL;
	}
	if (_OutDescr != NULL)
	{
		delete _OutDescr;
		_OutDescr = NULL;
	}
	_ResultSetAvailable = false;
	_Type = IBPP::stUnknown;
}

void StatementImpl::SetNull(int param)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetNull", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetNull", "The statement does not take parameters.");

	_InDescr->SetNull(param);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, bool value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetBool", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetBool", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivBool, &value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, const char* string)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetString", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetString", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivString, (void*)string);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, short value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetSmallInt", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetSmallInt", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivShort, &value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, int value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetInteger", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetInteger", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivLong, &value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, long value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetInteger", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetInteger", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivLong, &value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, __int64 value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetLargeInt", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetLargeInt", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivLarge, &value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, float value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetFloat", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetFloat", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivFloat, &value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, double value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetDouble", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetDouble", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivDouble, &value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, IBPP::Timestamp& value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetTimeStamp", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetTimeStamp", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivTimestamp, (void*)&value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, IBPP::Date& value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetDate", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetDate", "The statement does not take parameters.");

/*
	if (_Database->GetDialect() < 3)
	{
		// This is an IB 5 database. IDate is supposed to work with old 'DATE'
		// fields which are actually ISC_TIMESTAMP.
		ITimeStampImpl timestamp;
		timestamp.SetDate(value);
		_InDescr->SetValue(param, ivTimestamp, (void*)&timestamp);
	}
	else
	{
		// IB 6, dialect 3
*/
		_InDescr->SetValue(param, ivDate, (void*)&value);

/*
	}
*/

	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, IBPP::Time& value)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetTime", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetTime", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivTime, (void*)&value);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int param, IBPP::Blob* blob)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetBlob", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetBlob", "The statement does not take parameters.");
	if (dynamic_cast<DatabaseImpl*>(blob->GetDatabase()) != _Database)
		throw ExceptionImpl("Statement::SetBlob",
			"IBlob and Statement attached to different databases");
	if (dynamic_cast<TransactionImpl*>(blob->GetTransaction()) != _Transaction)
		throw ExceptionImpl("Statement::SetBlob",
			"IBlob and Statement attached to different transactions");

	_InDescr->SetValue(param, ivBlob, (void*)blob);
	_InMissing[param-1] = false;
}

void StatementImpl::Set(int column, IBPP::Array* array)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetBlob", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetBlob", "The statement does not take parameters.");

	_InDescr->SetValue(column, ivArray, (void*)array);
	_InMissing[column-1] = false;
}

void StatementImpl::Set(int param, IBPP::DBKey& key)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::SetDBKey", "No statement has been prepared.");
	if (_InDescr == NULL)
		throw ExceptionImpl("Statement::SetDBKey", "The statement does not take parameters.");

	_InDescr->SetValue(param, ivDBKey, (void*)&key);
	_InMissing[param-1] = false;
}

bool StatementImpl::IsNull(int column)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::IsNull", "The row is not initialized.");

	return _OutDescr->IsNull(column);
}

bool StatementImpl::Get(int column, bool* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == NULL)
		throw ExceptionImpl("Statement::Get", "NULL pointer detected");

	void* value = _OutDescr->GetValue(column, ivBool);
	if (value == NULL) *retvalue = false;
	else *retvalue = *(bool*)value;
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, char* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == NULL)
		throw ExceptionImpl("Statement::Get", "NULL pointer detected");

	void* value = _OutDescr->GetValue(column, ivString);
	if (value == NULL) strcpy(retvalue, "");
	else strcpy(retvalue, (char*)value);
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, short* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == NULL)
		throw ExceptionImpl("Statement::Get", "NULL pointer detected");

	void* value = _OutDescr->GetValue(column, ivShort);
	if (value == NULL) *retvalue = 0;
	else *retvalue = *(short*)value;
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, int* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == NULL)
		throw ExceptionImpl("Statement::Get", "NULL pointer detected");

	void* value = _OutDescr->GetValue(column, ivLong);
	if (value == NULL) *retvalue = 0;
	else *retvalue = *(int*)value;
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, long* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == NULL)
		throw ExceptionImpl("Statement::Get", "NULL pointer detected");

	void* value = _OutDescr->GetValue(column, ivLong);
	if (value == NULL) *retvalue = 0;
	else *retvalue = *(long*)value;
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, __int64* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == NULL)
		throw ExceptionImpl("Statement::Get", "NULL pointer detected");

	void* value = _OutDescr->GetValue(column, ivLarge);
	if (value == NULL) *retvalue = 0;
	else *retvalue = *(__int64*)value;
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, float* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == NULL)
		throw ExceptionImpl("Statement::Get", "NULL pointer detected");

	void* value = _OutDescr->GetValue(column, ivFloat);
	if (value == NULL) *retvalue = 0.0;
	else *retvalue = *(float*)value;
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, double* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == NULL)
		throw ExceptionImpl("Statement::Get", "NULL pointer detected");

	void* value = _OutDescr->GetValue(column, ivDouble);
	if (value == NULL) *retvalue = 0.0;
	else *retvalue = *(double*)value;
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, IBPP::Timestamp& timestamp)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* value = _OutDescr->GetValue(column, ivTimestamp, (void*)&timestamp);
	if (value == NULL) timestamp.Clear();
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, IBPP::Date& date)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

/*
	if (_Database->GetDialect() < 3)
	{
		// This is an IB 5 database. IDate is supposed to work with old 'DATE'
		// fields which are actually ISC_TIMESTAMP.
		ITimeStampImpl timestamp;
		void* value = _OutDescr->GetValue(column, ivTimestamp, (void*)&timestamp);
		if (null != NULL) *null = (value == NULL) ? true : false;
		if (value == NULL) date.Clear();
		else timestamp.GetDate(date);
	}
	else
	{
*/
		void* value = _OutDescr->GetValue(column, ivDate, (void*)&date);
		if (value == NULL) date.Clear();
		return value == NULL ? true : false;
/*
	}
*/
}

bool StatementImpl::Get(int column, IBPP::Time& time)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* value = _OutDescr->GetValue(column, ivTime, (void*)&time);
	if (value == NULL) time.Clear();
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, IBPP::Blob* retblob)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	retblob->AttachDatabase(_Database);
	retblob->AttachTransaction(_Transaction);
	void* value = _OutDescr->GetValue(column, ivBlob, (void*)retblob);
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, IBPP::DBKey& retkey)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* value = _OutDescr->GetValue(column, ivDBKey, (void*)&retkey);
	return value == NULL ? true : false;
}

bool StatementImpl::Get(int column, IBPP::Array* retarray)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* value = _OutDescr->GetValue(column, ivArray, (void*)retarray);
	return value == NULL ? true : false;
}

bool StatementImpl::IsNull(const char* name)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::IsNull", "The row is not initialized.");

	return IsNull(_OutDescr->GetVarNum(name));
}

bool StatementImpl::Get(const char* name, bool* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::GetBool", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, char* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::GetString", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, short* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, int* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, long* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, __int64* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, float* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, double* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, IBPP::Timestamp& retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, IBPP::Date& retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, IBPP::Time& retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, IBPP::Blob* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, IBPP::DBKey& retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const char* name, IBPP::Array* retvalue)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(_OutDescr->GetVarNum(name), retvalue);
}

int StatementImpl::Columns(void)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Columns", "The row is not initialized.");

	return _OutDescr->DescrSize();
}

int StatementImpl::ColumnNum(const char* name)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::ColumnNum", "The row is not initialized.");
	if (name == NULL)
		throw ExceptionImpl("Statement::ColumnNum", "NULL parameter detected.");

	return _OutDescr->GetVarNum(name);
}

const char* StatementImpl::ColumnName(int varnum)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Columns", "The row is not initialized.");

	return _OutDescr->GetName(varnum);
}

const char* StatementImpl::ColumnAlias(int varnum)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Columns", "The row is not initialized.");

	return _OutDescr->GetAlias(varnum);
}

const char* StatementImpl::ColumnTable(int varnum)
{
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::Columns", "The row is not initialized.");

	return _OutDescr->GetTable(varnum);
}

IBPP::SDT StatementImpl::ColumnType(int varnum)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::ColumnType", "No statement has been prepared.");
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::ColumnType", "The statement does not return results.");

    return _OutDescr->GetType(varnum);
}

int StatementImpl::ColumnSize(int varnum)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::ColumnSize", "No statement has been prepared.");
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::ColumnSize", "The row is not initialized.");

	return _OutDescr->GetSize(varnum);
}

int StatementImpl::ColumnScale(int varnum)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Statement::ColumnScale", "No statement has been prepared.");
	if (_OutDescr == NULL)
		throw ExceptionImpl("Statement::ColumnScale", "The row is not initialized.");

	return _OutDescr->GetScale(varnum);
}

void StatementImpl::Release(IBPP::Statement*& Self)
{
	if (this != dynamic_cast<StatementImpl*>(Self))
		throw ExceptionImpl("Statement::Release", "Invalid Release()");
	delete this;
	Self = NULL;
}

//	(((((((( OBJECT INTERNAL METHODS ))))))))

void StatementImpl::CursorFree(void)
{
	if (_ResultSetAvailable)
	{
		if (_Handle != NULL)
		{
			_IBS status;
			(*GDS::gds()->_dsql_free_statement)(status.Self(), &_Handle, DSQL_close);
			if (status.Errors())
				throw ExceptionImpl(&status, "StatementImpl::CursorFree",
					"isc_dsql_free_statement failed.");
		}
		_ResultSetAvailable = false;
	}
}

void StatementImpl::Init(void)
{
	_Handle = NULL;
	_Database = NULL;
	_Transaction = NULL;
	_InMissing = NULL;
	_InDescr = NULL;
	_OutDescr = NULL;
	_ResultSetAvailable = false;
	_Type = IBPP::stUnknown;
}

StatementImpl::StatementImpl(DatabaseImpl* database, TransactionImpl* transaction,
	const char* sql)
{
	Init();
	AttachDatabase(database);
	if (transaction != NULL) AttachTransaction(transaction);
	if (sql != NULL) Prepare(sql);
}

StatementImpl::~StatementImpl()
{
	try {Close();} catch (IBPP::Exception&) {}
	if (_Transaction != NULL) _Transaction->DetachStatement(this);
	if (_Database != NULL) _Database->DetachStatement(this);
}

//
//	EOF
//
