///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: test1.cpp,v 1.14 2001/11/10 11:18:25 epocman Exp $
//	Subject : IBPP, TEST1 program
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef IBPP_LINUX
#include <unistd.h>
#define DeleteFile(x) unlink(x)
#define __cdecl	/**/
#endif

//	The following database name (DbName) will be used during this test.
//	If existing, that DB will be deleted (assuming not used).
//	It will then be dynamically created, and various tests will run
//	while it is being populated and exercised.
//
//	The DB should grow to around 25 MB while being exercised.

#ifdef IBPP_LINUX
char* DbName = "/usr/tmp/TEST.FDB";
char* BkName = "/usr/tmp/TEST.FBK";
#else
char* DbName = "C:/TEST.FDB";	// FDB extension
								// GDB is hacked by Windows Me/XP "System Restore"
char* BkName = "C:/TEST.FBK";
#endif

//	The tests connect by default through the 'local' protocol to the server.
//	If you do not have an Interbase 6 server running locally on your machine,
//	or if you want to force it to use network protocol for the communication,
//	change the ServerName constant below to your server name. For instance :
//
//	char* ServerName = "localhost";

char* ServerName = NULL;

//	The tests use by default the well-known default of SYSDBA/masterkey
//	credentials. Do not forget to update them if required.

char* UserName = "SYSDBA";
char* Password = "masterkey";

class Test
{
	// Class 'Test' drives all the tests of this module.

	bool _Success;			// if true after all tests, everything succeeded
	char* _ProgramName;		// Will point to argv[1]

	void Test1(void);
	void Test2(void); 
	void Test3(void);
	void Test4(void);
	void Test5(void);
	void Test6(void);
	void Test7(void);

public:
	void RunTests(void);
	int PrintResult(void);
	Test(char* pname);
	~Test();
};

int Test::PrintResult(void)
{
	if (_Success) printf("\n*** SUCCESS ***\nAll tests were 100%% successfull.\n");
	else printf("\n*** FAILED ***\nSome tests failed, as indicated above.\n");
	return _Success ? 0 : 1;
}

void Test::RunTests(void)
{
	int NextTest = 1;

	printf("\nIBPP Test Suite (library version %d.%d.%d.%d)\n\n",
		(IBPP::Version & 0xFF000000) >> 24,
		(IBPP::Version & 0x00FF0000) >> 16,
		(IBPP::Version & 0x0000FF00) >> 8,
		(IBPP::Version & 0x000000FF));

	while (true)
	{
#ifdef IBPP_WINDOWS
		__try
		{
#endif
		try
		{
			switch (NextTest++)
			{
				case 1 :	Test1(); break;
				case 2 :	Test2(); break;
				case 3 :	Test3(); break;
				case 4 :	Test4(); break;
				case 5 :	Test5(); break;
				case 6 :	Test6(); break;
				case 7 :	Test7(); break;
				default :	return;		// All tests have been run
			}
		}
		catch (IBPP::Exception& e)
		{
			_Success = false;
			e.PrintfErrorMessage();
		}
		catch (...)
		{
			_Success = false;
			printf("Test %d --- FAILED\n"
				"A non-IBPP C++ runtime exception occured !\n\n", NextTest-1);
		}
#ifdef IBPP_WINDOWS
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			_Success = false;
			printf("Test %d --- FAILED\n"
				"A C-structured exception occured !\n\n", NextTest-1);
		}
#endif
	}
}

void Test::Test1(void)
{
	printf("Test 1 --- Checking Date / Time utility functions\n");

	IBPP::Date dt;
	IBPP::Date dt1;
	IBPP::Date dt2;
	int asint;

	dt.SetDate(2001, 2, 12);
	asint = dt;
	if (asint != 36933)
	{
		_Success = false;
		printf("Failed '12 February 2001' date as int test.\n"
			"Returned %d while 36933 was expected.\n", asint);
		return;
	}

	dt = asint;
	dt2 = dt;
	dt2.Add(15);
	if ((int)dt2 != 36948)
	{
		_Success = false;
		printf("Failed Date::Add() test.\n"
			"Returned %d while 36933 was expected.\n", (int)dt2);
		return;
	}

	int diff = dt2 - dt;
	if (diff != 15)
	{
		_Success = false;
		printf("Failed date arithmetic test.\n"
			"Returned %d while 15 was expected.\n", diff);
		return;
	}
}

void Test::Test2(void)
{
	printf("Test 2 --- Exercise empty database creation & connection\n");

	int Major, Minor, PageSize;
	bool Sync;

	IBPP::Database* db1;
	DeleteFile(DbName);
	db1 = IBPP::DatabaseFactory(ServerName, DbName, UserName, Password);
	db1->Create();	// Would fail if file existed, hence the previous DeleteFile
	db1->Connect();	// A create, does not imply connection

	db1->GetInfo(&Major, &Minor, &PageSize, &Sync);
	if (Sync)
	{
		_Success = false;
		printf("The created database has sync writes enabled,\n"
			"while it was not expected to.\n");
		return;
	}

	printf("ODS Major %d, ", Major);
	printf("ODS Minor %d, ", Minor);
	printf("Page Size %d, ", PageSize);
	printf("Sync Writes %s\n", Sync ? "true" : "false");

	db1->Disconnect();
	db1->Release(db1);
	if (db1 != NULL)
	{
		_Success = false;
		printf("Database::Release() not working.\n"
			"Object pointer should have been NULL on return.\n");
	}
}

void Test::Test3(void)
{
	printf("Test 3 --- Exercise basic DDL operations and IBPP::Exceptions\n");

	IBPP::Database* db1;
	db1 = IBPP::DatabaseFactory(ServerName, DbName, UserName, Password);
	db1->Connect();

	// The following transaction configuration values are the defaults and
	// those parameters could have as well be omitted to simplify writing.
	IBPP::Transaction* tr1 = IBPP::TransactionFactory(db1,
							IBPP::amWrite, IBPP::ilConcurrency, IBPP::lrWait);
	tr1->Start();

	IBPP::Statement* st1 = IBPP::StatementFactory(db1, tr1);

	st1->ExecuteImmediate(	"CREATE TABLE TEST("
							"N2 NUMERIC(9,2), "
							"N6 NUMERIC(15,2), "
							"N5 NUMERIC(9,5), "
							"D DATE, "
							"B BLOB SUB_TYPE 1, "
							"TF CHAR(1), "
							"ID INTEGER)");
	tr1->CommitRetain();
	st1->ExecuteImmediate(	"CREATE VIEW PRODUCT(X, Y) AS "
							"SELECT T.N2, S.N2 FROM TEST T, TEST S"
							);
	tr1->CommitRetain();
	printf("\n------------------------------------------------------------------------------\n");
	printf("The following exception is expected and is an integral part of the test suite.\n");
	printf("------------------------------------------------------------------------------\n");
	try
	{
		st1->ExecuteImmediate(	"CREATE SYNTAX ERROR(X, Y) AS "
								"SELECT ERRONEOUS FROM MUSTFAIL M" );
	}
	catch (IBPP::Exception& e)
	{
		e.PrintfErrorMessage();
	}
	printf("------------------------------------------------------\n");
	printf("End of the only expected exception of this test suite.\n");
	printf("------------------------------------------------------\n");

	// Intentionally do not Commit or Rollback the transaction, nor even Disconnect
	// The Release() will have to Rollback and terminate everything cleanly by itself.
	db1->Release(db1);
	if (db1 != NULL)
	{
		_Success = false;
		printf("Database::Release() not working.\n"
			"Object pointer should have been NULL on return.\n");
	}
	tr1->Release(tr1);
	if (tr1 != NULL)
	{
		_Success = false;
		printf("Transaction::Release() not working.\n"
			"Object pointer should have been NULL on return.\n");
	}
	st1->Release(st1);
	if (st1 != NULL)
	{
		_Success = false;
		printf("Statement::Release() not working.\n"
			"Object pointer should have been NULL on return.\n");
	}
}

void Test::Test4(void)
{
	printf("Test 4 --- Populate database and exercise Blobs (1000 rows)\n");

	IBPP::Database* db1;
	db1 = IBPP::DatabaseFactory(ServerName, DbName, UserName, Password);
	db1->Connect();

	// The following transaction configuration values are the defaults and
	// those parameters could have as well be omitted to simplify writing.
	IBPP::Transaction* tr1 = IBPP::TransactionFactory(db1,
							IBPP::amWrite, IBPP::ilConcurrency, IBPP::lrWait);
	tr1->Start();

	IBPP::Statement* st1 = IBPP::StatementFactory(db1, tr1);

	IBPP::Blob* b1 = IBPP::BlobFactory(db1, tr1);

	st1->Prepare("insert into test(N2,N6,N5,D,B,TF,ID) values(?, ?, ?, ?, ?, ?, ?)");
	st1->Set(1, 12.3456);
	st1->Set(2, 12.3456);
	st1->Set(3, 987.65432);
	IBPP::Date dt1;
	dt1.SetDate(2001, 2, 12);
	IBPP::Date dt2 = dt1;
	st1->Set(4, dt2);

	int i;
	int total = 0;
	for (i = 0; i < 1000; i++)
	{
		char buffer[1024];
		int len;
		b1->Create();
		FILE* file = fopen("ibpp.h", "r");
		while ((len = fread(buffer, 1, 1024, file)) == 1024)
		{
			b1->Write(buffer, 1024);
			total += 1024;
		}
		b1->Write(buffer, len);
		total += len;
		b1->Close();
		fclose(file);
		st1->Set(5, b1);
		st1->Set(6, (i%2) != 0);
		st1->Set(7, i+1);
		st1->Execute();
	}
	total /= 1000;

	b1->Release(b1);
	if (b1 != NULL)
	{
		_Success = false;
		printf("Blob::Release() not working.\n"
			"Object pointer should have been NULL on return.\n");
	}

	tr1->CommitRetain();

	st1->Prepare("select B from test where ID = 1");
	st1->Execute();
	st1->Fetch();

	IBPP::Blob* b2 = IBPP::BlobFactory(db1,tr1);
	char buffer[1024];
	int size, largest, segments;

	st1->Get(1, b2);
	b2->Open();
	b2->GetInfo(&size, &largest, &segments);
	if (size != total)
	{
		_Success = false;
		printf("Blob size incorrect on read back from the database.\n");
		printf("Read %d bytes, expected %d\n", size, total);
	}
	b2->Read(buffer, 1024);
	fflush(stdout);
	b2->Close();
	b2->Release(b2);

	st1->Close();	// unneeded : just to test
	tr1->Commit();
	db1->Disconnect();
	db1->Release(db1);
	tr1->Release(tr1);
	st1->Release(st1);
}

void Test::Test5(void)
{
	printf("Test 5 --- Cocktail of DML statements (1000 rows)\n");

	IBPP::Database* db1;
	db1 = IBPP::DatabaseFactory(ServerName, DbName, UserName, Password);
	db1->Connect();

	// The following transaction configuration values are the defaults and
	// those parameters could have as well be omitted to simplify writing.
	IBPP::Transaction* tr1 = IBPP::TransactionFactory(db1,
							IBPP::amWrite, IBPP::ilConcurrency, IBPP::lrWait);
	tr1->Start();

	IBPP::Statement* st1 = IBPP::StatementFactory(db1, tr1);

	IBPP::Blob* b1 = IBPP::BlobFactory(db1, tr1);

	st1->Prepare("select ID, rdb$db_key, N2, D, TF from TEST FOR UPDATE");
	st1->CursorExecute("MYCURSOR");
	/*
	for (int ii = 1; ii <= st1->Columns(); ii++)
	{
		printf("Column %s, Alias %s, Table %s, Type %d, Size %d, Scale %d\n",
			st1->ColumnName(ii), st1->ColumnAlias(ii), st1->ColumnTable(ii),
			(int)st1->ColumnType(ii), st1->ColumnSize(ii), st1->ColumnScale(ii));
	}
	*/

	IBPP::Statement* st2 = IBPP::StatementFactory(db1, tr1,
		"UPDATE TEST set n2 = ?, TF = ? where current of MYCURSOR");
	int incr = 0;
	while (st1->Fetch())
	{
		IBPP::Date d2;
		IBPP::Date d3;
		int y, m, d;
		long temp;
		st1->Get(4, d2);
		temp = d2;
		d3 = temp;
		d3 = d3 - incr;
		incr++;
		d3.GetDate(y, m, d);
		//printf("%d, %d, %d ",  y, m, d);
		double tmp;
		st1->Get(3, &tmp);
		st2->Set(1, tmp*40.0);
		bool b;
		st1->Get(5, &b);
		//printf("%s\n", b ? "true" : "false");
		/*
		printf("%s\n", row->Get("TF"));
		printf("%d\n", row->Get("ID"));
		*/
		st2->Set(2, ! b);
		st2->Execute();
	}
	st2->Release(st2);
	b1->Release(b1);

	st1->Prepare("select sum(N2),sum(N6) from test");
	st1->Execute();
	while (st1->Fetch())
	{
		double n2, n6;
		st1->Get(1, &n2);
		st1->Get(2, &n6);
		//double n2 = st1->GetDouble(1);
		//double n6 = st1->GetDouble(2);
		//printf("%g, %g\n", n2, n6);
	}

	//printf("Select returning no rows...\n");
	st1->Prepare("select N2 from test where N2 = 456278");
	st1->Execute();
	while (st1->Fetch())
	{
		_Success = false;
		printf("Statement::Fetch() not working.\n"
			"Returned row when there is none.\n");
		//double n2 = st1->GetDouble(1);
		//double n6 = st1->GetDouble(2);
		//printf("%g, %g\n", n2, n6);
	}

	//printf("Again, without prepare again...\n");
	st1->Execute();
	while (st1->Fetch())
	{
		_Success = false;
		printf("Statement::Fetch() not working.\n"
			"Returned row when there is none.\n");
		//double n2 = st1->GetDouble(1);
		//double n6 = st1->GetDouble(2);
		//printf("%g, %g\n", n2, n6);
	}


   //	printf("Executing 10 times an Execute, Prepared once...\n");
	st1->Prepare("select N2, N6 from test");
	for (int i = 0; i < 10; i++)
	{
		//printf("Loop %d\n", i);
		st1->Execute();
		while (st1->Fetch())
		{
			double n2, n6;
 			st1->Get(1, &n2);
			st1->Get(2, &n6);
			//double n2 = st1->GetDouble(1);
			//double n6 = st1->GetDouble(2);
			//printf("%g, %g\n", n2, n6);
		}
	}

	tr1->Commit();
	db1->Release(db1);
	tr1->Release(tr1);
	st1->Release(st1);
}

void Test::Test6(void)
{
	printf("Test 6 --- Service API (backup/restore/manage users)\n");

	IBPP::Service* svc = IBPP::ServiceFactory(ServerName, UserName, Password);
	svc->Connect();

	char msg[512];
	printf("Backup...\r");
	svc->StartBackup(DbName, BkName, true);
	//while (svc->Wait(msg)) printf("%s\n", msg);
	svc->Wait();
	DeleteFile(DbName);

	printf("Restore...\r");
	svc->StartRestore(BkName, DbName, 0, true);
	//while (svc->Wait(msg)) printf("%s\n", msg);
	svc->Wait();

	printf("Manage users...\r");
	svc->RemoveUser("OLIVIERM");
	svc->AddUser("OLIVIERM", "test", "Olivier", "Gilles", "Mascia");
	svc->ModifyUser("OLIVIERM", NULL, "Benoit", "", NULL);
	// On doit avoir un user OLIVIER, password test, Benoit rien Mascia

	printf("               \r");
	svc->Release(svc);
	if (svc != NULL)
	{
		_Success = false;
		printf("Service::Release() not working.\n"
			"Object pointer should have been NULL on return.\n");
	}
}

void Test::Test7(void)
{
	printf("Test 7 --- Mass delete, AffectedRows() and statistics\n");

	IBPP::Database* db1;
	db1 = IBPP::DatabaseFactory(ServerName, DbName, UserName, Password);
	db1->Connect();

	IBPP::Transaction* tr1 = IBPP::TransactionFactory(db1, IBPP::amWrite);
	tr1->Start();

	IBPP::Statement* st1 = IBPP::StatementFactory(db1, tr1);
	st1->Prepare("delete from test");
	st1->Execute();
	if (st1->AffectedRows() != 1000)
	{
		_Success = false;
		printf("Statement::AffectedRows() not working.\n"
			"Returned %d when 1000 was expected.\n", st1->AffectedRows());
	}
	else printf("Affected rows : %d\n", st1->AffectedRows());

	tr1->Commit();
	tr1->Release(tr1);
	st1->Release(st1);

	int Pages, Memory, Buffers, Fetches, Marks, Reads, Writes;

	db1->GetStat(&Pages, &Memory, &Buffers, &Fetches, &Marks, &Reads, &Writes);
	printf("Pages     : %d\n", Pages);
	printf("Memory    : %d\n", Memory);
	printf("Buffers   : %d\n", Buffers);
	printf("Fetches   : %d\n", Fetches);
	printf("Marks     : %d\n", Marks);
	printf("Reads     : %d\n", Reads);
	printf("Writes    : %d\n", Writes);

	db1->Release(db1);
}

Test::Test(char* pname)
{
	_ProgramName = pname;
	_Success = true;
}

Test::~Test()
{
}

int __cdecl main(int, char* argv[])
{
	Test T(argv[1]);

	if (! IBPP::CheckVersion(IBPP::Version))
	{
		printf("\nThis program got linked to an incompatible version of the library.\n"
			"Can't execute safely.\n");
		return 2;
	}

	T.RunTests();

	IBPP::Release();

	return T.PrintResult();
}

//	Eof
