///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: test2.cpp,v 1.7 2001/07/17 22:07:22 themule Exp $
//	Subject : IBPP, TEST1 program
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef IBPP_LINUX
#include <unistd.h>
#define DeleteFile(x) unlink(x)
#define Sleep(x) usleep(x)
#endif

//
//	The following database name (DbName) will be used during this test.
//	If existing, that DB will be crashed (assuming not used).
//	It will then be dynamically created, and various tests will run while
//	it is being populated and exercised.
//
//	DON'T FORGET TO CHANGE THAT DATABASE PATH AND NAME TO SOMETHING SUITABLE
//	TO YOUR CONFIGURATION !!
//
//	The tests connect by default through TCP/IP to the 'localhost'.
//

#ifdef IBPP_LINUX
char* ServerName=NULL;
char *DbName = "test.gdb";
#else
char* ServerName = "localhost";
char* DbName = "C:/Test.gdb";
#endif

void TestBanner(void)
{
	printf("IBPP, TEST PROGRAM '%s'\n", __FILE__);
	printf("EXERCISING 'EVENTS''\n\n");
}

class EventCatch : public IBPP::EventInterface
{
	virtual void ibppEventHandler(IBPP::Database*, const char* name, int count)
	{
		printf("Event %s triggered, count = %d\n", name, count);
	}
};

void run(void)
{
	if (! IBPP::CheckVersion(IBPP::Version))
	{
		printf("This program got linked to an incompatible version of the library.\n"
			"Can't execute safely.\n");
		return;
	}

	EventCatch catcher;

	IBPP::Database* db1;
	printf("Creating a fresh new database...\n");
	DeleteFile(DbName);
	db1 = IBPP::DatabaseFactory(ServerName, DbName,	"SYSDBA", "masterkey");
	db1->Create();	// Would fail if file existed
	db1->Connect();	// A create, does not imply connection

	// The following transaction configuration values are the defaults and
	// those parameters could have as well be omitted to simplify writing.
	IBPP::Transaction* tr1 = IBPP::TransactionFactory(db1,
							IBPP::amWrite, IBPP::ilConcurrency, IBPP::lrWait);
	tr1->Start();

	IBPP::Statement* st1 = IBPP::StatementFactory(db1, tr1);
	printf("Creating schema...\n");
	st1->ExecuteImmediate(	"CREATE TABLE TEST("
							"N1 NUMERIC(18,2), "
							"N2 NUMERIC(18,2) "
							")");
	tr1->CommitRetain();
	st1->ExecuteImmediate(
        "CREATE TRIGGER TEST_TRIGGER FOR TEST ACTIVE AFTER INSERT AS\n"
        "BEGIN\n"
		"	POST_EVENT 'INSERT';\n"
        "END"   );

	tr1->Commit();

	db1->DefineEvent("FIRST", &catcher);
	db1->DefineEvent("SECOND", &catcher);
	db1->DefineEvent("INSERT", &catcher);	// The only one we use
	//db1->DropEvent("SECOND");
	printf("First Dispatch Events\n");
	db1->DispatchEvents();

	printf("Sleeping 2 sec...\n");
	Sleep(2000);

	printf("Adding an event\n");
	db1->DefineEvent("FOURTH", &catcher);

	printf("Second Dispatch\n");
	db1->DispatchEvents();

	printf("Inserting 2 records, that should trigger 'INSERT' event...\n");
	tr1->Start();
	st1->ExecuteImmediate("INSERT INTO TEST(N1) VALUES(1)");
	st1->ExecuteImmediate("INSERT INTO TEST(N1) VALUES(1)");

	printf("Third Dispatch (commit not done...\n");
	db1->DispatchEvents();

	printf("Now committing\n");
	tr1->Commit();

	printf("Series of 20 calls to Dispatch, with a 0.050 sec sleep after each\n");
	for (int i = 0; i < 20; i++)
	{
		db1->DispatchEvents();
		Sleep(50);
	}
	db1->Release(db1);

	IBPP::Release();
}

int main(int, char* [])
{
	TestBanner();

	try	{ run(); }
		catch (IBPP::Exception& e) { e.PrintfErrorMessage(); }

	return 0;
}

//
//	Eof
//
