///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: time.cpp,v 1.4 2001/03/04 00:26:27 themule Exp $
//	Subject : IBPP, Time class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

void IBPP::Time::Now(void)
{
	SYSTEMTIME st;
	GetLocalTime(&st);

	IBPP::itot(&_Time, (int)st.wHour, (int)st.wMinute, (int)st.wSecond);
}

void IBPP::Time::SetTime(int hour, int minute, int second)
{
	if (hour < 0 || hour > 23 ||
		minute < 0 || minute > 59 ||
			second < 0 || second > 59)
				throw ExceptionHelper("Time::SetTime",
					"Invalid hour, minute, second values");
	IBPP::itot(&_Time, hour, minute, second);
}

void IBPP::Time::GetTime(int& hour, int& minute, int& second)
{
	IBPP::ttoi(_Time, &hour, &minute, &second);
}

IBPP::Time::Time(int hour, int minute, int second)
{
	SetTime(hour, minute, second);
}

IBPP::Time::Time(const IBPP::Time& copied)
{
	_Time = copied._Time;
}

IBPP::Time::Time(const int& copied)
{
	if (copied < 0 || copied > 86399) throw ExceptionHelper("Time::Time(int&)", "Invalid time value");
	_Time = copied;
}

IBPP::Time& IBPP::Time::operator=(const IBPP::Time& assigned)
{
	_Time = assigned._Time;
	return *this;
}

IBPP::Time& IBPP::Time::operator=(const int& assigned)
{
	if (assigned < 0 || assigned > 86399)
		throw ExceptionHelper("Time::operator=(int)", "Invalid time value");
	_Time = assigned;
	return *this;
}

//
//	EOF
//

