///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: timestamp.cpp,v 1.5 2001/11/03 11:13:17 epocman Exp $
//	Subject : IBPP, Timestamp class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

void IBPP::Timestamp::Today(void)
{
	SYSTEMTIME st;
	GetLocalTime(&st);
	IBPP::itod(&_Date, (int)st.wYear, (int)st.wMonth, (int)st.wDay);
	_Time = 0;
}

void IBPP::Timestamp::Now(void)
{
	SYSTEMTIME st;
	GetLocalTime(&st);
	IBPP::itod(&_Date, (int)st.wYear, (int)st.wMonth, (int)st.wDay);
	IBPP::itot(&_Time, (int)st.wHour, (int)st.wMinute, (int)st.wSecond);
}

void IBPP::Timestamp::SetDate(int year, int month, int day)
{
	IBPP::itod(&_Date, year, month, day);
	if (_Date < 1) throw ExceptionHelper("Timestamp::SetDate", "Invalid year, month, day values");
}

void IBPP::Timestamp::GetDate(int& year, int& month, int& day)
{
	IBPP::dtoi(_Date, &year, &month, &day);
}

void IBPP::Timestamp::SetTime(int hour, int minute, int second)
{
	if (hour < 0 || hour > 23 ||
		minute < 0 || minute > 59 ||
			second < 0 || second > 59)
				throw ExceptionHelper("Timestamp::SetTime",
					"Invalid hour, minute, second values");
	IBPP::itot(&_Time, hour, minute, second);
}

void IBPP::Timestamp::GetTime(int& hour, int& minute, int& second)
{
	IBPP::ttoi(_Time, &hour, &minute, &second);
}

void IBPP::Timestamp::Add(int days)
{
	int year;
	int newdate = _Date + days;		// days can be signed
	IBPP::dtoi(newdate, &year, NULL, NULL);
	if (year == 0) throw ExceptionHelper("Timestamp::Add()", "Invalid resulting date");
	_Date = newdate;
}

void IBPP::Timestamp::StartOfMonth(void)
{
	int year, month;
	IBPP::dtoi(_Date, &year, &month, NULL);
	IBPP::itod(&_Date, year, month, 1);		// First of same month
	_Time = 0;	// Midnight
}

void IBPP::Timestamp::EndOfMonth(void)
{
	int year, month;
	IBPP::dtoi(_Date, &year, &month, NULL);
	if (++month > 12) { month = 1; year++; }
	IBPP::itod(&_Date, year, month, 1);		// First of next month
	_Date--;	// Last day of original month, all weird cases accounted for
	IBPP::itot(&_Time, 23, 59, 59);		// Last second of day
}

IBPP::Timestamp::Timestamp(int year, int month, int day)
{
	SetDate(year, month, day);
	_Time = 0;
}

IBPP::Timestamp::Timestamp(int year, int month, int day,
	int hour, int minute, int second)
{
	SetDate(year, month, day);
	SetTime(hour, minute, second);
}

IBPP::Timestamp::Timestamp(const IBPP::Timestamp& copied)
{
	_Date = copied._Date;
	_Time = copied._Time;
}

IBPP::Timestamp& IBPP::Timestamp::operator=(const IBPP::Timestamp& assigned)
{
	_Date = assigned._Date;
	_Time = assigned._Time;
	return *this;
}

IBPP::Timestamp& IBPP::Timestamp::operator=(const IBPP::Date& assigned)
{
	_Date = (IBPP::Date)assigned;	// Use the operator=(int) (cast is drop const)
	return *this;
}

IBPP::Timestamp& IBPP::Timestamp::operator=(const IBPP::Time& assigned)
{
	_Time = (IBPP::Time)assigned;	// Use the operator=(int) (cast is drop const)
	return *this;
}

//
//	EOF
//
