///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: transaction.cpp,v 1.9 2001/11/10 12:52:58 epocman Exp $
//	Subject : IBPP, Database class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

//	(((((((( OBJECT INTERFACE IMPLEMENTATION ))))))))

void TransactionImpl::AttachDatabase(IBPP::Database* db,
	IBPP::TAM am, IBPP::TIL il, IBPP::TLR lr)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Transaction::AttachDatabase",
				"Can't attach a Database if Transaction started.");
	if (db == NULL)
		throw ExceptionImpl("Transaction::AttachDatabase",
				"Can't attach a null Database.");

	// Alloue de nouveaux tableaux
	DatabaseImpl** newDatabases = new DatabaseImpl* [_DbCount+1];
	_TPB** newTPBs = new _TPB* [_DbCount+1];

	// Copie les tableaux existants (s'il y en avait)
	if (_DbCount > 0)
	{
		memcpy((BYTE*)newDatabases, (BYTE*)_Databases, _DbCount*sizeof(DatabaseImpl*));
		memcpy((BYTE*)newTPBs, (BYTE*)_TPBs, _DbCount*sizeof(_TPB*));
	}

	// Ajoute les nouveaux lments
	newDatabases[_DbCount] = dynamic_cast<DatabaseImpl*>(db);
	newTPBs[_DbCount] = new _TPB();
	_TPB* tpb = newTPBs[_DbCount];

	// Prpare le nouveau _TPB
    if (am == IBPP::amRead) tpb->Insert(isc_tpb_read);
    else tpb->Insert(isc_tpb_write);

	switch (il)
	{
		case IBPP::ilConsistency :	tpb->Insert(isc_tpb_consistency); break;
		case IBPP::ilReadDirty :		tpb->Insert(isc_tpb_read_committed);
						    	    tpb->Insert(isc_tpb_rec_version); break;
		case IBPP::ilReadCommitted :	tpb->Insert(isc_tpb_read_committed);
									tpb->Insert(isc_tpb_no_rec_version); break;
		default :				tpb->Insert(isc_tpb_concurrency); break;
	}

    if (lr == IBPP::lrNoWait) tpb->Insert(isc_tpb_nowait);
    else tpb->Insert(isc_tpb_wait);

	// Supprime les tableaux existants (s'il y en avait)
	if (_Databases != NULL)	delete [] _Databases;
	if (_TPBs != NULL) delete [] _TPBs;

	// Attache les nouveaux tableaux
	_Databases = newDatabases;
	_TPBs = newTPBs;

	// Incrmente le nombre d'lments
	_DbCount++;

	// Informe l'objet Database qu'il a bien t ajout  la transaction
	db->AttachTransaction(this);
}

void TransactionImpl::DetachDatabase(IBPP::Database* db)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Transaction::DetachDatabase",
				"Can't detach a Database if Transaction started.");
	if (db == NULL)
		throw ExceptionImpl("Transaction::DetachDatabase",
				"Can't detach a null Database.");

	for (int i = 0; i < _DbCount; i++)
	{
		ASSERTION(_Databases != NULL)
		ASSERTION(_Databases[i] != NULL)
		if (_Databases[i] == dynamic_cast<DatabaseImpl*>(db))
		{
			// La Database dont il faut se dtacher est identifie.
			// Eliminons-la du tableau par dcalage des autres lments.
			// Il importe peu qu'un lment allou devienne inutilis. Si on
			// attache une autre Database  cette transaction, de nouveaux
			// tableaux seront de toute faon allous.
			ASSERTION(_TPBs != NULL)
			ASSERTION(_TPBs[i] != NULL)
			delete _TPBs[i];
			for (int j = i; j < _DbCount-1; j++)
			{
				_Databases[j] = _Databases[j+1];
				_TPBs[j] = _TPBs[j+1];
			}
			_DbCount--;
			ASSERTION(_DbCount >= 0)
			break;
		}
	}
}

void TransactionImpl::Start(void)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Transaction::Start", "Transaction already started.");
	if (_DbCount == 0)
		throw ExceptionImpl("Transaction::Start", "No Database is attached.");

    _IBS status;
	struct ISC_TEB
	{
		long* db_ptr;
		long tpb_len;
		char* tpb_ptr;
	} * teb = new ISC_TEB[_DbCount];
    int i;

    for (i = 0; i < _DbCount; i++)
    {
		if (_Databases[i]->GetHandle() == NULL)
		{
			// Toutes les Database doivent etre connectees prealablement !
			delete [] teb;
			throw ExceptionImpl("Transaction::Start",
					"All attached Database should have been started.");
		}
		teb[i].db_ptr = (long*) _Databases[i]->GetHandlePtr();
		teb[i].tpb_len = _TPBs[i]->Size();
		teb[i].tpb_ptr = _TPBs[i]->Self();
	}

    (*GDS::gds()->_start_multiple)(status.Self(), &_Handle, (short)_DbCount, teb);
    delete [] teb;
    if (status.Errors())
    {
        _Handle = NULL;     // Should be, but better be sure...
		throw ExceptionImpl(&status, "Transaction::Start");
    }
}

void TransactionImpl::Commit(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Transaction::Commit", "Transaction is not started.");
		
    _IBS status;

	(*GDS::gds()->_commit_transaction)(status.Self(), &_Handle);
    if (status.Errors())
		throw ExceptionImpl(&status, "Transaction::Commit");

	int i;
	for (i = _StCount-1; i >= 0; i--)
		try { _Statements[i]->CursorFree(); }
			catch (IBPP::Exception&) { }

    _Handle = NULL;
}

void TransactionImpl::CommitRetain(void)
{	  
	if (_Handle == NULL)
		throw ExceptionImpl("Transaction::CommitRetain", "Transaction is not started.");

    _IBS status;

    (*GDS::gds()->_commit_retaining)(status.Self(), &_Handle);
    if (status.Errors())
		throw ExceptionImpl(&status, "Transaction::CommitRetain");
}

void TransactionImpl::Rollback(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Transaction::Rollback", "Transaction is not started.");

    _IBS status;

    (*GDS::gds()->_rollback_transaction)(status.Self(), &_Handle);
    if (status.Errors())
		throw ExceptionImpl(&status, "Transaction::Rollback");

	int i;
	for (i = _StCount-1; i >= 0; i--)
		try { _Statements[i]->CursorFree(); }
			catch (IBPP::Exception&) { }

    _Handle = NULL;
}

void TransactionImpl::RollbackRetain(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Transaction::RollbackRetain", "Transaction is not started.");

    _IBS status;

    (*GDS::gds()->_rollback_retaining)(status.Self(), &_Handle);
    if (status.Errors())
		throw ExceptionImpl(&status, "Transaction::RollbackRetain");
}

void TransactionImpl::Release(IBPP::Transaction*& Self)
{
	if (this != dynamic_cast<TransactionImpl*>(Self))
		throw ExceptionImpl("Transaction::Release", "Invalid Release()");
	delete this;
	Self = NULL;
}

//	(((((((( OBJECT INTERNAL METHODS ))))))))

void TransactionImpl::Init(void)
{
	_Handle = NULL;
	_DbCount = 0; _Databases = NULL; _TPBs = NULL;
	_StCount = 0; _Statements = NULL;
 	_BbCount = 0; _Blobs = NULL;
	_ArCount = 0; _Arrays = NULL;
}

void TransactionImpl::AttachStatement(StatementImpl* st)
{
	if (st == NULL)
		throw ExceptionImpl("Transaction::AttachStatement",
					"Can't attach a NULL Statement object.");

	// Alloue un nouveau tableau
	StatementImpl** newStatements = new StatementImpl* [_StCount+1];
	// Recopie l'ancien, s'il existait
	if (_StCount > 0)
		memcpy((BYTE*)newStatements, (BYTE*)_Statements,
			_StCount*sizeof(StatementImpl*));
	// Ajoute le nouvel lment
	newStatements[_StCount] = st;
	// Dtruit l'ancien tableau
	if (_Statements != NULL) delete [] _Statements;
	// Affecte le nouveau tableau
	_Statements = newStatements;
	// Incrmente le nombre de transactions associes
	_StCount++;
		
	return;
}

void TransactionImpl::DetachStatement(StatementImpl* st)
{
	if (st == NULL)
		throw ExceptionImpl("Transaction::DetachStatement",
				"Can't detach a NULL Statement object.");

	// Recherche ce statement
	for (int i = 0; i < _StCount; i++)
	{
		ASSERTION(_Statements != NULL)
		ASSERTION(_Statements[i] != NULL)
		if (_Statements[i] == st)
		{
			// Le statement dont il faut se dtacher est identifi.
			// Eliminons-la du tableau par dcalage des autres lments.
			// Il importe peu qu'un lment allou devienne inutilis. Si on
			// attache Database  un autre statement, un nouveau tableau
			// sera de toute faon allou.
			for (int j = i; j < _StCount-1; j++)
				_Statements[j] = _Statements[j+1];
			_StCount--;
			ASSERTION(_StCount >= 0)
			break;
		}
	}
}

void TransactionImpl::AttachBlob(BlobImpl* bb)
{
	if (bb == NULL)
		throw ExceptionImpl("Transaction::AttachBlob",
					"Can't attach a NULL BlobImpl object.");

	// Alloue un nouveau tableau
	BlobImpl** newBlobs = new BlobImpl* [_BbCount+1];
	// Recopie l'ancien, s'il existait
	if (_BbCount > 0)
		memcpy((BYTE*)newBlobs, (BYTE*)_Blobs,
			_BbCount*sizeof(BlobImpl*));
	// Ajoute le nouvel lment
	newBlobs[_BbCount] = bb;
	// Dtruit l'ancien tableau
	if (_Blobs != NULL)	delete [] _Blobs;
	// Affecte le nouveau tableau
	_Blobs = newBlobs;
	// Incrmente le nombre de blobs attachs
	_BbCount++;
		
	return;
}

void TransactionImpl::DetachBlob(BlobImpl* bb)
{
	if (bb == NULL)
		throw ExceptionImpl("Transaction::DetachBlob",
				"Can't detach a NULL BlobImpl object.");

	// Recherche ce Blob
	for (int i = 0; i < _BbCount; i++)
	{
		ASSERTION(_Blobs != NULL)
		ASSERTION(_Blobs[i] != NULL)
		if (_Blobs[i] == bb)
		{
			// Le Blob dont il faut se dtacher est identifi.
			// Eliminons-le du tableau par dcalage des autres lments.
			// Il importe peu qu'un lment allou devienne inutilis. Si on
			// attache Transaction  un autre Blob, un nouveau tableau
			// sera de toute faon allou.
			for (int j = i; j < _BbCount-1; j++)
				_Blobs[j] = _Blobs[j+1];
			_BbCount--;
			ASSERTION(_BbCount >= 0)
			break;
		}
	}
}

void TransactionImpl::AttachArray(ArrayImpl* ar)
{
	if (ar == NULL)
		throw ExceptionImpl("Transaction::AttachArray",
					"Can't attach a NULL ArrayImpl object.");

	// Alloue un nouveau tableau
	ArrayImpl** newArrays = new ArrayImpl* [_ArCount+1];
	// Recopie l'ancien, s'il existait
	if (_ArCount > 0)
		memcpy((BYTE*)newArrays, (BYTE*)_Arrays,
			_ArCount*sizeof(ArrayImpl*));
	// Ajoute le nouvel lment
	newArrays[_ArCount] = ar;
	// Dtruit l'ancien tableau
	if (_Arrays != NULL) delete [] _Arrays;
	// Affecte le nouveau tableau
	_Arrays = newArrays;
	// Incrmente le nombre de Arrays attachs
	_ArCount++;
		
	return;
}

void TransactionImpl::DetachArray(ArrayImpl* ar)
{
	if (ar == NULL)
		throw ExceptionImpl("Transaction::DetachArray",
				"Can't detach a NULL ArrayImpl object.");

	// Recherche ce Array
	for (int i = 0; i < _ArCount; i++)
	{
		ASSERTION(_Arrays != NULL)
		ASSERTION(_Arrays[i] != NULL)
		if (_Arrays[i] == ar)
		{
			// Le Array dont il faut se dtacher est identifi.
			// Eliminons-le du tableau par dcalage des autres lments.
			// Il importe peu qu'un lment allou devienne inutilis. Si on
			// attache Transaction  un autre Array, un nouveau tableau
			// sera de toute faon allou.
			for (int j = i; j < _ArCount-1; j++)
				_Arrays[j] = _Arrays[j+1];
			_ArCount--;
			ASSERTION(_ArCount >= 0)
			break;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
//	INTERNAL (HIDDEN) IMPLEMENTATION
//	PUBLIC METHODS
///////////////////////////////////////////////////////////////////////////////

TransactionImpl::TransactionImpl(DatabaseImpl* db,
	IBPP::TAM am, IBPP::TIL il, IBPP::TLR lr)
{
	Init();
	AttachDatabase(db, am, il, lr);
}

TransactionImpl::~TransactionImpl()
{
	// Abandonnons la transaction si elle tait dmarre
	if (_Handle != NULL)
	{
		try { Rollback(); } catch (IBPP::Exception&) { }
	}

	// Dtachons cet Transaction de tout BlobImpl. C'est trs important afin
	// que les objets BlobImpl n'aient plus de rfrence  un Transaction qui
	// n'existe plus.
	// Balayage inverse du tableau pour viter effets de bords et amliorer performance.
	// En effet le tableau rtrci  chaque tour de boucle (_BbCount diminue).
	// Et lors de la suppression, il y compactage du tableau par copie d'lments
	// du haut vers le bas.
	int i;
	for (i = _BbCount-1; i >= 0; i--)
		try { _Blobs[i]->DetachTransaction(); }
			catch (IBPP::Exception&) { }

	// Dtachons cet Transaction de tout Array. C'est trs important afin
	// que les objets ArrayImpl n'aient plus de rfrence  un Transaction qui
	// n'existe plus.
  	for (i = _ArCount-1; i >= 0; i--)
		try { _Arrays[i]->DetachTransaction(); }
			catch (IBPP::Exception&) { }

	// Dtachons cet Transaction de tout Statement. C'est trs important afin
	// que les objets Statement n'aient plus de rfrence  un Transaction qui
	// n'existe plus.
	for (i = _StCount-1; i >= 0; i--)
		try { _Statements[i]->DetachTransaction(); }
			catch (IBPP::Exception&) { }

	// Trs important : dtachons proprement toutes les bases de donnes de
	// cette transaction. Il ne faut plus que des objets Database aient encore
	// un pointeur vers cet Transaction qui disparait.
    for (i = _DbCount-1; i >= 0; i--)
    	try { _Databases[i]->DetachTransaction(this); }
    		catch (IBPP::Exception&) { }

	// Finalement, libration des tableaux
	if (_Arrays != NULL) delete [] _Arrays;
	if (_Blobs != NULL) delete [] _Blobs;
	if (_Statements != NULL) delete [] _Statements;
	if (_Databases != NULL) delete [] _Databases;
	if (_TPBs != NULL) delete [] _TPBs;
}

//
//	EOF
//
